from unittest import TestCase
from mapactionpy_controller.main_stack import process_stack
from mapactionpy_controller.steps import Step
import logging
import random
from time import sleep
# works differently for python 2.7 and python 3.x
try:
    from unittest import mock
except ImportError:
    import mock


class TestSteps(TestCase):

    @mock.patch('humanfriendly.terminal.connected_to_terminal')
    def test_get_args(self, mock_hft):
        # mock_hft is purely to prevent the console from becoming too clutered when running unittests
        mock_hft.return_value = False

        process_stack(self._get_demo_steps(secs=0.1), None)
        self.assertTrue(True)

    def _get_demo_steps(self, secs=3):
        def random_pass(**kwargs):
            sleep(secs)
            if random.random() > 0.5:
                raise ValueError('Something went wrong')

            return 'stopped for {} secs'.format(secs)

        demo_steps = [
            Step(
                random_pass,
                logging.WARNING,
                'DEMO: Checking that the Crash Move Folder description file opens correctly',
                'DEMO: The Crash Move Folder description file opened correctly',
                'DEMO: Failed to open the Crash Move Folder description file correctly',
            ),
            Step(
                random_pass,
                logging.WARNING,
                'DEMO: Checking that each of the configuration files matches their relevant schemas',
                'DEMO: Each of the configuration files adheres to their relevant schemas',
                'DEMO: Failed to verify one or more of the configuration files against the relevant schema',
            ),
            Step(
                random_pass,
                logging.WARNING,
                'DEMO: Comparing the contents of the layer properties json file and the layer rendering directory',
                'DEMO: Compared the contents of the layer properties json file and the layer rendering directory',
                'DEMO: Inconsistancy found in between the contents of the layer properties json file and the layer'
                ' rendering directory'
            ),
            Step(
                random_pass,
                logging.WARNING,
                'DEMO: Comparing the contents of the layer properties json file and the MapCookbook',
                'DEMO: Compared the contents of the layer properties json file and the MapCookbook',
                'DEMO: Inconsistancy found in between the contents of the layer properties json file'
                ' and the MapCookbook'
            )
        ]

        return demo_steps




import glob
import logging
import os

import mapactionpy_controller.name_convention as name_convention
from mapactionpy_controller.crash_move_folder import CrashMoveFolder
from mapactionpy_controller.event import Event
from mapactionpy_controller.steps import Step
from mapactionpy_controller.task_renderer import FixDataNameTask
from mapactionpy_controller.data_search import get_all_gisfiles


def get_defaultcmf_step_list(cmf_config_path):
    """
    Generates a list of Steps, each of which execucutes a suitable naming convention test
    for a static item within the crash move folder.
    """
    cmf = CrashMoveFolder(cmf_config_path)

    ncs_to_check = (
        (cmf.layer_rendering, cmf.layer_nc_definition, '.lyr', 'layer'),
        (cmf.layer_rendering, cmf.layer_nc_definition, '.qml', 'layer'),
        (cmf.layer_rendering, cmf.layer_nc_definition, '.qlr', 'layer'),
        (cmf.map_projects, cmf.map_projects_nc_definition, '.qgs', 'map project'),
        (cmf.map_projects, cmf.map_projects_nc_definition, '.mxd', 'map project'),
        (cmf.map_templates, cmf.map_template_nc_definition, '.qgs', 'map template'),
        (cmf.map_templates, cmf.map_template_nc_definition, '.aprx', 'map template'),
        (cmf.map_templates, cmf.map_template_nc_definition, '.pagx', 'map template'),
        (cmf.map_templates, cmf.map_template_nc_definition, '.mxd', 'map template')
    )

    name_convention_steps = []

    for dir_to_check, nc_desc_file, extn_to_check, convention_name in ncs_to_check:
        nc = name_convention.NamingConvention(nc_desc_file)
        name_convention_steps.extend(
            _step_builer(_get_all_files(dir_to_check, extn_to_check), nc, convention_name, cmf))

    return name_convention_steps


def _get_all_files(dir, extn):
    return glob.glob('{}/*{}'.format(dir, extn))


def get_active_data_step_list(humev_config_path):
    """
    Generates a list of Steps, each of which execucutes a suitable naming convention test
    for a GIS dataset with the `2_Active_data` folder within the crash move folder.
    """
    humev = Event(humev_config_path)
    cmf = CrashMoveFolder(humev.cmf_descriptor_path)
    nc = name_convention.NamingConvention(cmf.data_nc_definition)
    return _step_builer(get_all_gisfiles(cmf), nc, 'data', cmf)


def get_single_file_checker(f_path, nc, cmf):
    def check_data_name(**kwargs):
        f_name = os.path.basename(f_path)
        ncr = nc.validate(f_name)
        if not ncr.is_valid:
            raise ValueError(FixDataNameTask(ncr, cmf))

        # https://trello.com/c/BODjWZrw/184-implement-check-that-gis-files-are-in-the-right-folder
        # Check that the file is in the right directory to go here
        return ncr

    return check_data_name


def _step_builer(file_list, nc, convention_name, cmf):
    step_list = []
    for f_path in file_list:
        # return_code += check_dir(dir_to_check, nc_desc_file, extn_to_check, args.inc_valid)
        base_name = os.path.basename(f_path)
        step_list.append(
            Step(
                get_single_file_checker(f_path, nc, cmf),
                logging.WARNING,
                "Checking the file '{}' against the {} naming convention".format(base_name, convention_name),
                "The file '{}' matches the {} naming convention".format(base_name, convention_name),
                "The file '{}' does not match the {} naming convention".format(base_name, convention_name)
            )
        )

    return step_list



import logging
from unittest import TestCase
from mapactionpy_controller.main_stack import process_stack
from mapactionpy_controller.steps import Step
try:
    from unittest import mock
except ImportError:
    import mock


class TestMainStack(TestCase):

    def get_append_state(self, step_name, state_obj_type):
        def append_to_list(**kwargs):
            state = kwargs['state']
            state.append(step_name)
            return state

        def append_to_dict(**kwargs):
            state = kwargs['state']
            state[step_name] = step_name
            return state

        def append_to_string(**kwargs):
            state = kwargs['state']
            state = state + step_name
            return state

        func_dict = {
            'list': append_to_list,
            'dict': append_to_dict,
            'string': append_to_string
        }

        return func_dict[state_obj_type]

    def build_example_steps(self, str_list, state_obj_type):
        step_list = []
        for step_name in str_list:
            step_list.append(
                Step(
                    self.get_append_state(step_name, state_obj_type),
                    logging.WARNING,
                    'DEMO: Doing  {}'.format(step_name),
                    'DEMO: Passed {}'.format(step_name),
                    'DEMO: Failed {}'.format(step_name)
                )
            )

        return step_list

    def setUp(self):
        pass

    @mock.patch('mapactionpy_controller.main_stack.hft.connected_to_terminal')
    def test_simple_step_list(self, mock_hft):
        # mock_hft is purely to prevent the console from becoming too clutered when running unittests
        mock_hft.return_value = False

        starter_list = ['AAA', 'BBB', 'CCC', 'DDD']

        test_cases = [
            ('list', [], ['AAA', 'BBB', 'CCC', 'DDD']),
            ('dict', {}, {'AAA': 'AAA', 'BBB': 'BBB', 'CCC': 'CCC', 'DDD': 'DDD'}),
            ('string', '', 'AAABBBCCCDDD')
        ]

        for state_type, initial_state, expected_state in test_cases:
            starter_steps = self.build_example_steps(starter_list, state_type)
            final_state = process_stack(starter_steps, initial_state)
            self.assertEqual(expected_state, final_state)
            # print('\nfinal state = {}'.format(final_state))

    @mock.patch('mapactionpy_controller.main_stack.hft.connected_to_terminal')
    def test_add_single_step_into_stack(self, mock_hft):
        # mock_hft is purely to prevent the console from becoming too clutered when running unittests
        mock_hft.return_value = False

        starter_list = ['AAA', 'BBB', 'CCC', 'DDD']

        test_cases = [
            ('list', [], ['AAA', 'BBB', '111', 'DDD']),
            ('dict', {}, {'AAA': 'AAA', 'BBB': 'BBB', '111': '111', 'DDD': 'DDD'}),
            ('string', '', 'AAABBB111DDD')
        ]

        for state_type, initial_state, expected_state in test_cases:
            def get_inject_steps(**kwargs):
                return self.build_example_steps(['111'], state_type)[0]

            starter_steps = self.build_example_steps(starter_list, state_type)
            # change the func on one of the step, to a function which returns more Step objects
            starter_steps[2].func = get_inject_steps
            final_state = process_stack(starter_steps, initial_state)
            self.assertEqual(expected_state, final_state)
            # print('\nfinal list = {}'.format(final_state))

    @mock.patch('mapactionpy_controller.main_stack.hft.connected_to_terminal')
    def test_add_multiple_steps_into_stack(self, mock_hft):
        # mock_hft is purely to prevent the console from becoming too clutered when running unittests
        mock_hft.return_value = False

        starter_list = ['AAA', 'BBB', 'CCC', 'DDD']

        test_cases = [
            ('list', [], ['AAA', 'BBB', '111', '222', '333', 'DDD']),
            ('dict', {}, {'AAA': 'AAA', 'BBB': 'BBB', '111': '111', '222': '222', '333': '333', 'DDD': 'DDD'}),
            ('string', '', 'AAABBB111222333DDD')
        ]

        for state_type, initial_state, expected_state in test_cases:
            def get_inject_steps(**kwargs):
                return self.build_example_steps(['111', '222', '333'], state_type)

            starter_steps = self.build_example_steps(starter_list, state_type)
            # change the func on one of the step, to a function which returns more Step objects
            starter_steps[2].func = get_inject_steps
            final_state = process_stack(starter_steps, initial_state)
            self.assertEqual(expected_state, final_state)
            # print('\nfinal list = {}'.format(final_state))



import unittest
import mapactionpy_controller.data_search as data_search
import mapactionpy_controller.recipe_layer as recipe_layer
from mapactionpy_controller.map_recipe import MapRecipe
from mapactionpy_controller.event import Event
from mapactionpy_controller.layer_properties import LayerProperties
from mapactionpy_controller.crash_move_folder import CrashMoveFolder
import fixtures
import os
import six
import platform


# works differently for python 2.7 and python 3.x
if six.PY2:
    import mock  # noqa: F401
    # from mock import mock_open, patch
else:
    from unittest import mock  # noqa: F401
    # from unittest.mock import mock_open, patch  # noqa: F401


class TestDataSearch(unittest.TestCase):

    def setUp(self):
        self.parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.cmf_descriptor_path = os.path.join(self.parent_dir, 'example', 'cmf_description_flat_test.json')
        self.event_descriptor_path = os.path.join(self.parent_dir, 'example', 'event_description.json')
        self.recipe_file = os.path.join(self.parent_dir, 'example', 'example_single_map_recipe.json')
        self.non_existant_file = os.path.join(self.parent_dir, 'example', 'non-existant-file.json')
        self.cmf = CrashMoveFolder(self.cmf_descriptor_path, verify_on_creation=False)
        self.event = Event(self.event_descriptor_path)
        # self.cmf = CrashMoveFolder(self.cmf_descriptor_path)
        self.lyr_props = LayerProperties(self.cmf, '', verify_on_creation=False)
        # self.ds = data_search.DataSearch(cmf_descriptor_path)

    def test_substitute_fields_in_recipe_strings(self):
        recipe_updater = data_search.get_recipe_event_updater(self.event)

        # An example with an iso code within a positive regex lookup
        # (eg '^{e.affected_country_iso3}_stle.....')
        reference_recipe = MapRecipe(
            fixtures.recipe_with_positive_iso3_code, self.lyr_props)
        pos_recipe = MapRecipe(
            fixtures.recipe_without_positive_iso3_code, self.lyr_props)
        updated_pos_recipe = recipe_updater(state=pos_recipe)
        self.assertEqual(updated_pos_recipe, reference_recipe)

        # An example with an iso code within a negitive regex lookup
        # (eg '^(?!({e.affected_country_iso3}))_admn.....')
        reference_recipe = MapRecipe(
            fixtures.recipe_with_negative_iso3_code, self.lyr_props)
        neg_recipe = MapRecipe(
            fixtures.recipe_without_negative_iso3_code, self.lyr_props)
        recipe_updater = data_search.get_recipe_event_updater(self.event)
        updated_neg_recipe = recipe_updater(state=neg_recipe)

        self.assertEqual(updated_neg_recipe, reference_recipe)

        # Reverse of the first example to test a case where nothing needs updating. In this case the same json
        # definition is used for both recipes. There are not string replacement fields within the json definition.
        reference_recipe = MapRecipe(
            fixtures.recipe_with_positive_iso3_code, self.lyr_props)
        pos_recipe = MapRecipe(
            fixtures.recipe_with_positive_iso3_code, self.lyr_props)
        updated_pos_recipe = recipe_updater(state=pos_recipe)
        self.assertEqual(updated_pos_recipe, reference_recipe)

    def test_search_for_shapefiles(self):
        # ds = data_search.DataSearch(self.event)

        with mock.patch('mapactionpy_controller.data_search.glob.glob') as mock_glob:
            # We use two different mock values becuase we are matching absolute paths as strings
            # The underlying production code does not differencate between platforms.
            if platform.system() == 'Windows':
                mock_single_file_glob = fixtures.glob_single_stle_file_search_windows
                mock_multiple_file_glob = fixtures.glob_multiple_stle_file_search_windows
                mock_no_file_glob = fixtures.glob_no_stle_file_search_windows
                ref_recipe_str = fixtures.recipe_result_one_dataset_per_layer_windows
            else:
                mock_single_file_glob = fixtures.glob_single_stle_file_search_linux
                mock_multiple_file_glob = fixtures.glob_multiple_stle_file_search_linux
                mock_no_file_glob = fixtures.glob_no_stle_file_search_linux
                ref_recipe_str = fixtures.recipe_result_one_dataset_per_layer_linux

            reference_recipe = MapRecipe(ref_recipe_str, self.lyr_props)

            # Case A
            # where there is exactly one dataset per query
            test_recipe = MapRecipe(fixtures.recipe_test_for_search_for_shapefiles, self.lyr_props)
            mock_glob.return_value = mock_single_file_glob
            all_gis_files = [(f_path, os.path.basename(f_path)) for f_path in data_search.get_all_gisfiles(self.cmf)]
            self.assertNotEqual(test_recipe, reference_recipe)

            # get the first layer from the test_recipe
            test_lyr = test_recipe.all_layers().pop()
            data_finder = test_lyr.get_data_finder(self.cmf, all_gis_files)
            updated_test_recipe = data_finder(state=test_recipe)

            self.assertEqual(updated_test_recipe, test_recipe)
            self.assertTrue(updated_test_recipe == test_recipe)
            self.assertEqual(updated_test_recipe, reference_recipe)

            # Case B
            # where there is multiple matching datasets
            test_recipe = MapRecipe(fixtures.recipe_test_for_search_for_shapefiles, self.lyr_props)
            mock_glob.return_value = mock_multiple_file_glob
            all_gis_files = [(f_path, os.path.basename(f_path)) for f_path in data_search.get_all_gisfiles(self.cmf)]
            self.assertNotEqual(test_recipe, reference_recipe)

            # get the first layer from the test_recipe
            test_lyr = test_recipe.all_layers().pop()
            data_finder = test_lyr.get_data_finder(self.cmf, all_gis_files)
            with self.assertRaises(ValueError) as arcm:
                updated_test_recipe = data_finder(state=test_recipe)

            ve = arcm.exception
            print('ve.args[0] is instance of: {}'.format(type(ve.args[0])))
            self.assertIsInstance(ve.args[0], recipe_layer.FixMultipleMatchingFilesTask)

            # Case C
            # where there are no matching datasets
            test_recipe = MapRecipe(fixtures.recipe_test_for_search_for_shapefiles, self.lyr_props)
            mock_glob.return_value = mock_no_file_glob
            all_gis_files = [(f_path, os.path.basename(f_path)) for f_path in data_search.get_all_gisfiles(self.cmf)]
            self.assertNotEqual(test_recipe, reference_recipe)

            # get the first layer from the test_recipe
            test_lyr = test_recipe.all_layers().pop()
            data_finder = test_lyr.get_data_finder(self.cmf, all_gis_files)
            with self.assertRaises(ValueError) as arcm:
                updated_test_recipe = data_finder(state=test_recipe)

            ve = arcm.exception
            print('ve.args[0] is instance of: {}'.format(type(ve.args[0])))
            self.assertIsInstance(ve.args[0], recipe_layer.FixMissingGISDataTask)

            # Case D
            # Case insensitive search. Like Case A, but we force the regex itself into upper and lower case.
            test_recipe = MapRecipe(fixtures.recipe_test_for_search_for_shapefiles, self.lyr_props)
            mock_glob.return_value = mock_single_file_glob
            all_gis_files = [(f_path, os.path.basename(f_path)) for f_path in data_search.get_all_gisfiles(self.cmf)]
            self.assertNotEqual(test_recipe, reference_recipe)

            # get the first layer from the test_recipe
            test_lyr = test_recipe.all_layers().pop()
            for new_reg_exp in [test_lyr.reg_exp.upper(), test_lyr.reg_exp.lower()]:
                # save the old Reg Exp in order to use to compare recipes later
                old_reg_exp = test_lyr.reg_exp
                test_lyr.reg_exp = new_reg_exp
                data_finder = test_lyr.get_data_finder(self.cmf, all_gis_files)
                updated_test_recipe = data_finder(state=test_recipe)
                # Now replace the origional reg_exp
                test_lyr.reg_exp = old_reg_exp

                self.assertEqual(updated_test_recipe, test_recipe)
                self.assertTrue(updated_test_recipe == test_recipe)
                self.assertEqual(updated_test_recipe, reference_recipe)

            # Case E
            # Check whether or not it is enforced that a regex is matches the beginning of the string or at any
            # point in the string. Simular to Case A
            test_recipe = MapRecipe(fixtures.recipe_test_for_search_for_shapefiles, self.lyr_props)
            mock_glob.return_value = mock_single_file_glob
            all_gis_files = [(f_path, os.path.basename(f_path)) for f_path in data_search.get_all_gisfiles(self.cmf)]
            self.assertNotEqual(test_recipe, reference_recipe)

            # get the first layer from the test_recipe
            test_lyr = test_recipe.all_layers().pop()
            # save the old Reg Exp in order to use to compare recipes later
            old_reg_exp = test_lyr.reg_exp
            # force an regex that doesn't have an anchor at the begining of the string, nor does it match the
            # beginning of the filename
            test_lyr.reg_exp = 'stle_ste_pt_(.*?)_(.*?)_([phm][phm])(.*?)'

            data_finder = test_lyr.get_data_finder(self.cmf, all_gis_files)
            updated_test_recipe = data_finder(state=test_recipe)

            # Now replace the origional reg_exp
            test_lyr.reg_exp = old_reg_exp

            self.assertEqual(updated_test_recipe, test_recipe)
            self.assertTrue(updated_test_recipe == test_recipe)
            self.assertEqual(updated_test_recipe, reference_recipe)

            # Case F
            # Check case where filesystem has case sensitive filenames, meaning that we could match on two
            # seperate files whose names only different by case.
            # Very simular to Case B
            test_recipe = MapRecipe(fixtures.recipe_test_for_search_for_shapefiles, self.lyr_props)
            mock_glob.return_value = fixtures.glob_multiple_stle_file_search_case_difference_linux
            all_gis_files = [(f_path, os.path.basename(f_path)) for f_path in data_search.get_all_gisfiles(self.cmf)]
            self.assertNotEqual(test_recipe, reference_recipe)

            # get the first layer from the test_recipe
            test_lyr = test_recipe.all_layers().pop()
            data_finder = test_lyr.get_data_finder(self.cmf, all_gis_files)
            with self.assertRaises(ValueError) as arcm:
                updated_test_recipe = data_finder(state=test_recipe)

            ve = arcm.exception
            self.assertIsInstance(ve.args[0], recipe_layer.FixMultipleMatchingFilesTask)

    def test_check_layer(self):

        recipe = MapRecipe(fixtures.recipe_with_layer_details_embedded, self.lyr_props)
        lyr = recipe.all_layers().pop()

        # This will pass silently if the lyr is of the right type
        self.assertIsNone(data_search._check_layer(lyr))

        # This case should raise a ValueError
        with self.assertRaises(ValueError):
            data_search._check_layer('string-to-represent-a-layer')



import logging

from mapactionpy_controller.event import Event
from mapactionpy_controller.layer_properties import LayerProperties
from mapactionpy_controller.map_cookbook import MapCookbook
from mapactionpy_controller.steps import Step
import mapactionpy_controller.data_search as data_search

# logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def get_plugin_step():
    def get_plugin(**kwargs):
        hum_event = kwargs['state']
        try:
            logger.debug('Attempting to load the ArcMapRunner')
            from mapactionpy_arcmap.arcmap_runner import ArcMapRunner
            runner = ArcMapRunner(hum_event)
            logger.info('Successfully loaded the ArcMapRunner')
        except ImportError:
            logger.debug('Failed to load the ArcMapRunner')
            logger.debug('Attempting to load the QGisRunner')
            from mapactionpy_qgis.qgis_runner import QGisRunner
            runner = QGisRunner()
            logger.info('Failed to load the ArcMapRunner')

        return runner

    def new_event(**kwargs):
        return Event(kwargs['state'])

    plugin_step = [
        Step(
            new_event,
            logging.ERROR,
            'Loading the Humanitarian Event description file',
            'Successfully loaded the Humanitarian Event description file',
            'Failed to load the Humanitarian Event description file',
        ),
        Step(
            get_plugin,
            logging.ERROR,
            'Identifying available plugins (ArcMapRunner/QGisRunner)',
            'Successfully loaded an available plugin',
            'Failed to load a suitable plugin',
        ),
    ]

    return plugin_step


def _get_product_start_step(recipe):
    def just_return_recipe(**kwargs):
        return recipe

    return Step(
        just_return_recipe,
        logging.ERROR,
        'Starting to create map "{}" - "{}"'.format(recipe.mapnumber, recipe.product),
        'Starting to create map "{}" - "{}"'.format(recipe.mapnumber, recipe.product),
        'Failed to create map "{}" - "{}"'.format(recipe.mapnumber, recipe.product),
    )


def _get_product_end_step(recipe):

    def pass_through_step(**kwargs):
        # recipe = kwargs['state']
        # print(recipe)
        pass

    return Step(
        pass_through_step,
        logging.ERROR,
        'Completed the creation of map "{}" - "{}"'.format(recipe.mapnumber, recipe.product),
        'Completed the creation of map "{}" - "{}"'.format(recipe.mapnumber, recipe.product),
        'Failed to create map "{}" - "{}"'.format(recipe.mapnumber, recipe.product),
    )


def _get_per_product_runner_steps(_runner, recipe):
    # In due course there should be greater granularity for some of these steps
    logger.debug('Building steps for recipe {}'.format(recipe.mapnumber))

    product_steps = [
        Step(
            _runner.get_templates,
            logging.ERROR,
            'Identifying suitable map template',
            'Successfully indentifed suitable map template',
            'Failed to identify suitable map template',
        ),
        Step(
            _runner.create_ouput_map_project,
            logging.ERROR,
            "Creating new '{}' file.".format(_runner.get_projectfile_extension()),
            "Successfully created new '{}' file.".format(_runner.get_projectfile_extension()),
            "Failed to create new '{}' file.".format(_runner.get_projectfile_extension())
        ),
        Step(
            _runner.build_project_files,
            logging.ERROR,
            'Adding layers to the map and applying styling',
            'Successfully added layers to the map and applying styling',
            'Failed to add the layers to the map and applied styling',
        ),
        Step(
            _runner.export_maps,
            logging.WARNING,
            'Exporting Maps and creating zipfile',
            'Successfully exported Maps and creating zipfile',
            'Failed to export the maps and create zipfile'
        )
    ]

    temp_msg = product_steps[0].running_msg
    logger.debug('Built steps for recipe {} with running_msg = {}'.format(recipe.mapnumber, temp_msg))

    return product_steps


def get_cookbook_steps(my_runner, map_number, dry_run, verify_on_creation=True):
    def get_cookbook(**kwargs):
        lyrs = LayerProperties(my_runner.cmf, my_runner.get_lyr_render_extension(), verify_on_creation)
        my_cookbook = MapCookbook(my_runner.cmf, lyrs, verify_on_creation, my_runner.hum_event)

        selected_product_steps = []
        for recipe in select_recipes(my_cookbook, map_number):
            logger.debug('About to create steps for recipe {}'.format(recipe.mapnumber))
            selected_product_steps.append(_get_product_start_step(recipe))
            selected_product_steps.extend(data_search.get_per_product_data_search_steps(
                my_runner, recipe))

            # This is a crude implenmentaiton of dry-run for now.
            if not dry_run:
                selected_product_steps.extend(_get_per_product_runner_steps(my_runner, recipe))

            selected_product_steps.append(_get_product_end_step(recipe))

        return selected_product_steps

    cookbook_steps = [
        Step(
            get_cookbook,
            logging.ERROR,
            'Openning the MapCookbook files',
            'Successfully opened the MapCookbook files',
            'Failed to open the MapCookbook files'
        )
    ]

    return cookbook_steps


def select_recipes(cookbook, map_nums=None):
    all_recipes = cookbook.products.values()

    if map_nums:
        try:
            cleaned_nums = [map_nums.upper()]
        except AttributeError:
            cleaned_nums = [mn.upper() for mn in map_nums]

        selected_recipes = [r for r in all_recipes if r.mapnumber.upper() in cleaned_nums]

        logger.debug('MapIDs "{}" have been selected'.format([r.mapnumber for r in selected_recipes]))
        return selected_recipes
    else:
        return all_recipes



import os
import logging
import six
from unittest import TestCase

from mapactionpy_controller import jira_tasks
from mapactionpy_controller.task_renderer import TaskReferralBase

# works differently for python 2.7 and python 3.x
if six.PY2:
    import mock  # noqa: F401
else:
    from unittest import mock  # noqa: F401


class TestJiraClient(TestCase):
    def setUp(self):
        self.parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.dir_to_valid_cmf_des = os.path.join(self.parent_dir, 'example')
        self.path_to_valid_cmf_des = os.path.join(self.dir_to_valid_cmf_des, 'cmf_description_flat_test.json')

    def test_jira_client_constructor(self):
        # Test without .netrc present
        fail_msg = 'Unable to locate or load suitable `.netrc` file for JIRA integration'
        mock_env = mock.patch.dict(os.environ, {"MAPCHEF_NETRC": "file-that-does-not-exist"})
        mock_env.start()
        print('env-var = {}'.format(os.environ["MAPCHEF_NETRC"]))
        self.call_jira_constrcutor(fail_msg)
        mock_env.stop()

        # Test with .netrc present but without matching machinename
        fail_msg = 'Unable to find details for machine'
        test_failing_netrc_path = os.path.join(self.parent_dir, 'tests', 'testfiles', 'test_netrc_wrong_machine_name')
        print(test_failing_netrc_path)
        mock_env = mock.patch.dict(os.environ, {"MAPCHEF_NETRC": test_failing_netrc_path})
        mock_env.start()
        print('env-var = {}'.format(os.environ["MAPCHEF_NETRC"]))
        self.call_jira_constrcutor(fail_msg)
        mock_env.stop()

        # Test with .netrc present but without correct authentication details
        with mock.patch('mapactionpy_controller.jira_tasks.JIRA'):
            fail_msg = 'Unable to authenticate with JIRA'
            test_failing_netrc_path = os.path.join(self.parent_dir, 'tests', 'testfiles', 'test_netrc_failing_auth')
            mock_env = mock.patch.dict(os.environ, {"MAPCHEF_NETRC": test_failing_netrc_path})
            mock_env.start()
            self.call_jira_constrcutor(fail_msg)
            mock_env.stop()

    def call_jira_constrcutor(self, fail_msg):
        with self.assertRaises(ValueError) as ve:
            jira_tasks.JiraClient()

        if six.PY2:
            self.assertRegexpMatches(str(ve.exception), fail_msg)
        else:
            self.assertRegex(str(ve.exception), fail_msg)

    @mock.patch('mapactionpy_controller.jira_tasks._check_jira_con')
    @mock.patch('mapactionpy_controller.jira_tasks.JIRA')
    def test_search_issue_by_unique_summary(self, mock_JIRA, mock_check_jira_con):
        test_failing_netrc_path = os.path.join(self.parent_dir, 'tests', 'testfiles', 'test_netrc_failing_auth')
        with mock.patch.dict(os.environ, {"MAPCHEF_NETRC": test_failing_netrc_path}):
            # This overrides JiraClient's own self check that it has a valid connection
            mock_check_jira_con.return_value = None
            jira_client = jira_tasks.JiraClient()
            test_op_id = 'testopid123'

            # test case were extact one task returned
            single_result = ['a']
            jira_client.jira_con.search_issues = mock.MagicMock(return_value=single_result)
            self.assertEqual('a', jira_client.search_issue_by_unique_summary('serach-string', test_op_id))

            # test case where no tasks returned
            jira_client.jira_con.search_issues = mock.MagicMock(return_value=None)
            self.assertIsNone(jira_client.search_issue_by_unique_summary('serach-string', test_op_id))
            jira_client.jira_con.search_issues = mock.MagicMock(return_value=[])
            self.assertIsNone(jira_client.search_issue_by_unique_summary('serach-string', test_op_id))

            # test case where multiple task returned
            multi_result = ['a', 'b']
            jira_client.jira_con.search_issues = mock.MagicMock(return_value=multi_result)

            with self.assertRaises(ValueError) as ve:
                fail_msg = 'More than one JIRA Issue found'
                jira_client.search_issue_by_unique_summary('serach-string', test_op_id)

                if six.PY2:
                    self.assertRegexpMatches(str(ve.exception), fail_msg)
                else:
                    self.assertRegex(str(ve.exception), fail_msg)

    @mock.patch('mapactionpy_controller.jira_tasks._check_jira_con')
    @mock.patch('mapactionpy_controller.jira_tasks.JIRA')
    def test_task_handler(self, mock_JIRA, mock_check_jira_con):
        test_failing_netrc_path = os.path.join(self.parent_dir, 'tests', 'testfiles', 'test_netrc_failing_auth')
        with mock.patch.dict(os.environ, {"MAPCHEF_NETRC": test_failing_netrc_path}):
            # This overrides JiraClient's own self check that it has a valid connection
            mock_check_jira_con.return_value = None
            jira_client = jira_tasks.JiraClient()

            # Case 1)
            # where there is no task object, therefore should pass sliently
            jira_client.jira_con.search_issues = mock.MagicMock(return_value=None)
            # This isn't a particularly diagnostic test, but in this case litterialy nothing happens
            self.assertIsNone(jira_client.task_handler(logging.ERROR, 'test-case', None))

            # Case 2)
            # where there is no existing issue for the task, therefore a new JIRA issue should be created
            jira_client.jira_con.search_issues = mock.MagicMock(return_value=None)
            with mock.patch.object(jira_client.jira_con, 'create_issue', return_value=None) as mock_method:
                test_task = TaskReferralBase(None)
                jira_client.task_handler(logging.ERROR, 'test-case', test_task)

            # assert that jira_client.create_new_jira_issue is called once
            mock_method.assert_called_once()

            # Case 3)
            # where there was an existing *open* task, which is in the correct column, that requires commenting on.
            test_task = TaskReferralBase(None)

            mock_issue = mock.Mock(name='mock_issue')
            mock_issue.id = "123"  # issue ID
            mock_issue.fields.summary = test_task.get_task_unique_summary()
            mock_issue.fields.description = test_task.get_task_description()
            mock_issue.fields.status.id = jira_client.target_column

            jira_client.jira_con.search_issues = mock.MagicMock(return_value=[mock_issue])
            with mock.patch.object(jira_client.jira_con, 'add_comment', return_value=None) as mock_add_comment_method:
                with mock.patch.object(jira_client.jira_con, 'update', return_value=None) as mock_update_method:
                    jira_client.task_handler(logging.ERROR, 'test-case', test_task)

            mock_update_method.assert_not_called()
            mock_add_comment_method.assert_called_once()
            comment_arg = mock_add_comment_method.call_args[0][1]

            target_msg = 'This Issue was still current'
            if six.PY2:
                self.assertRegexpMatches(comment_arg, target_msg)
            else:
                self.assertRegex(comment_arg, target_msg)

            # Case 4)
            # where the issue description requires updating
            test_task = TaskReferralBase(None)

            mock_issue = mock.Mock(name='mock_issue')
            mock_issue.id = "123"  # issue ID
            mock_issue.fields.summary = test_task.get_task_unique_summary()
            # Gaurentee that the description for the issue is different to the task.
            mock_issue.fields.description = 'description is different. {}'.format(test_task.get_task_description())
            mock_issue.fields.status.id = jira_client.target_column

            jira_client.jira_con.search_issues = mock.MagicMock(return_value=[mock_issue])
            with mock.patch.object(jira_client.jira_con, 'add_comment', return_value=None) as mock_add_comment_method:
                with mock.patch.object(mock_issue, 'update', return_value=None) as mock_update_method:
                    jira_client.task_handler(logging.ERROR, 'test-case', test_task)

            mock_update_method.assert_called_once()
            mock_add_comment_method.assert_called_once()

            # Case 5)
            # where there was an existing *closed* task, that requires *re-opening* and commenting on.

            # Case 6)
            # where there is an no-longer relevant *open* task, that requires (possibly closing) an commenting on



# flake8: noqa
from unittest import TestCase

import mapactionpy_controller.data_source
import mapactionpy_controller.map_doc
import mapactionpy_controller.map_report
import mapactionpy_controller.map_result
import mapactionpy_controller.plugin_base
import mapactionpy_controller.plugin_controller


class TestAllStuffTransfered(TestCase):

    def setUp(self):
        pass

    def test_all_stuff_transfered(self):
        self.assertTrue(True)



import glob
import itertools
import logging
import os

from mapactionpy_controller.recipe_layer import RecipeLayer
from mapactionpy_controller.steps import Step


def get_recipe_event_updater(hum_event):
    """
    Creates a function which will update a recipe with situation specific information about the
    Humanitarian Event. Certain strings within a recipe can be can include "replacement fields"
    using Python's String format Syntax
    (https://docs.python.org/3.3/library/string.html#formatstrings)

    This method replaces those strings with runtime values. Replacement fields referencing the humanitarian
    event object should use the name `e`. For example, the value of `recipe.summary` could be set in
    the recipe files to:
    ```
    recipe.summary = "Overview map of {e.country_name}"
    ```
    This method will update that string with the country name from `self.event`:
    ```
    recipe.summary = "Overview map of Atlantis"
    ```

    The following fields within a recipe are updated:

    * `product`
    * `summary`
    * `lyr.reg_exp` for every layer
    * `lyr.definition_query`
    * `lbl_class.expression`
    * `lbl_class.sql_query`

    If the input strings do not include any replacement fields the recipe is returned unaltered.

    @kwargs state: The recipe to be updated.
    @returns The same recipe object with the relevant strings updated as appropriate.
    """
    def update_recipe_item(item):
        try:
            return item.format(e=hum_event)
        except IndexError:
            return item

    def update_recipe_with_event_details(**kwargs):
        recipe = kwargs['state']
        return _update_items_in_recipe(recipe, update_recipe_item)

    return update_recipe_with_event_details


def _update_items_in_recipe(recipe, update_recipe_item):
    for lyr in itertools.chain(*[mf.layers for mf in recipe.map_frames]):
        lyr.reg_exp = update_recipe_item(lyr.reg_exp)
        lyr.definition_query = update_recipe_item(lyr.definition_query)

        for lbl_class in lyr.label_classes:
            lbl_class.expression = update_recipe_item(lbl_class.expression)
            lbl_class.sql_query = update_recipe_item(lbl_class.sql_query)

    # Update the recipe level members
    recipe.product = update_recipe_item(recipe.product)
    recipe.summary = update_recipe_item(recipe.summary)

    return recipe


def _check_layer(recipe_lyr):
    """
    Checks that the recipe is a MapRecipe object.

    @param recipe_lyr: The object to be checked.
    @raises ValueError: If recipe_lyr is not an instance of MapRecipe.
    """
    if not isinstance(recipe_lyr, RecipeLayer):
        error_msg = (
            'Unable to `get_lyr_data_finder` when MapRecipe is not validated against LayerProerties.'
            ' This method needs a RecipeLayer object not just a string placeholder.'
            ' layer_name = "{}"'.format(recipe_lyr)
        )
        logging.error(error_msg)
        raise ValueError(error_msg)


def get_all_gisfiles(cmf):
    gisfiles_with_paths = set()

    for extn in ['.shp', '.img', '.tif']:
        for f_path in glob.glob('{}/*/*{}'.format(cmf.active_data, extn)):
            gisfiles_with_paths.add(f_path)

    return gisfiles_with_paths


def get_per_product_data_search_steps(runner, recipe):
    """

    1) Find all possible datasources for a layer
            JIRA if there are multiple datasources per layer

    2) Check that all datasources match the required schema
            JIRA if there is a schema mismatch.

    3) Calculate checksum

    4) Calculate extent

    """
    # ds = DataSearch(hum_event)
    all_gis_files = [(f_path, os.path.basename(f_path)) for f_path in get_all_gisfiles(runner.cmf)]

    step_list = [
        Step(
            get_recipe_event_updater(runner.hum_event),
            logging.ERROR,
            'Updating recipe with event specific details',
            'Updated recipe with event specific details',
            'Failed to update recipe with event specific details'
        )]

    for map_frame in recipe.map_frames:
        for recipe_lyr in map_frame.layers:
            # This is just being paraniod. This can only occur if somewhere up the chain a MapCookbook was created
            # with the param verify_on_creation=False. That should not occur in production code and only in
            # some test cases.
            _check_layer(recipe_lyr)
            step_list.extend([
                Step(
                    recipe_lyr.get_data_finder(runner.cmf, all_gis_files),
                    logging.WARNING,
                    '"Searching for data suitable for layer "{}"'.format(recipe_lyr.name),
                    'Found data for layer "{}"'.format(recipe_lyr.name),
                    'Failed to find data suitable for layer "{}"'.format(recipe_lyr.name)
                ),
                Step(
                    recipe_lyr.calc_extent,
                    logging.WARNING,
                    'Calculating extent for the data for layer "{}"'.format(recipe_lyr.name),
                    'Calculated extent for the data for layer "{}"'.format(recipe_lyr.name),
                    'Error whilst calculating extent for the data for layer "{}"'.format(recipe_lyr.name)
                ),
                Step(
                    recipe_lyr.check_data_against_schema,
                    logging.WARNING,
                    'Checking the schema for the data available for layer "{}"'.format(recipe_lyr.name),
                    'Verified schema the data available for layer "{}"'.format(recipe_lyr.name),
                    'The data available for layer "{}" failed schema check'.format(recipe_lyr.name)
                )
            ])

        step_list.extend([
            Step(
                map_frame.calc_extent,
                logging.WARNING,
                'Calculating extent for the map frame "{}"'.format(map_frame.name),
                'Calculated extent for the map frame "{}"'.format(map_frame.name),
                'Error whilst extent for the map frame "{}"'.format(map_frame.name)
            )
        ])

    return step_list



import os

import chevron

import mapactionpy_controller.name_convention as name_convention
from mapactionpy_controller import TASK_TEMPLATES_DIR


"""
Module `task_renderer`

This module provides as way to generate human-readable, context specifc task descriptions for any task
which MapChef needs to delegate to human to complete. The main class `TaskReferralBase` represents a catch-all
generic error condition. It is not expected to be used directly, but more commonly subclasses to represent
more specific senarios.

For the convenience of those creating subclasses of `TaskReferralBase`, there are some "adapter" functions
available, which can be used extract key information from common objects from within `mapactionpy_controller`
into a format suitable for `self.context_data`. They are particular useful for the constructors of any
sub-classes.

The adapter functions only cater for a single object of each type.
```
context_data = {}
context_data.expend(adapter1.(foo))
context_data.expend(adapter2.(bar))
```

Each adapter function maps an object to a key within `context_data`:
* `_name_result_adapter`; NamingResult -> 'name_result'
* `_cmf_description_adapter`; CrashMoveFolder -> `cmf`

"""


class TaskReferralBase(object):
    """
    Represents a senario where a task would need to be referred to a human. The class provides a unique
    identifer for the task, which must be useful by the external task-tracking system and the bolierplate
    description of the underlying problem that needs to be resolved.

    Different senarios are represented by subclassing this class. For example:
    * `DataFileNameTask` : A category of problems which relate to filenames.
    * `FixSchemaErrorTask` : A category of problems which relate to errors in the schema.

    Specific instances of thes object represent specific problems that require human intervention:
    * 'the file "foo.txt" is incorrectly named'
    * 'the file "bar.shp" is missing the attribute "place_name"'

    Subclasses must:
    A) Either:
        * Override the `_task_template_filename`. This should be the name of a file (excluding the extension).
          This file (which must have the extension `.mustache`) must exist in `TASK_TEMPLATES_DIR`.
        * Overide `_get_task_template` so that it returns the contents of a mustache template. This option
          allows for template files which exist in locations other than TASK_TEMPLATES_DIR, and hence can to
          more appropriate for plugins.
    B) Override `_primary_key_template` with a string mustache.io template as a string. When rendered template
       should provide a unique identifier for an instance of the class. This unique identifier must be valid
       (and meaningful) within the external task management system.
    C) Set `self.context_data` from within the constructor. `self.context_data` is a dict object that contains
       the relevant centextual information to be able to render the templates. The keys of `context_data` must
       match the tags within the mustache template for both the `_primary_key_template and the
       `_task_template_filename`. There are some adapter functions to help with this. These adapters take
       common objects from within `mapactionpy_controller` and extract key information into a format suitable
       for `self.context_data`.

    Note:
    The templates use the mustache.io format. However it is assumed the delimters left=`<%` and right=`%>` are
    used, rather than the eponymous defaults `{{` and `}}`. This is aviod ambiguity with JIRA's markup which
    also assigns special meaning to double-curly brackets.
    """
    _primary_key_template = 'Major Configuration Error'
    _task_template_filename = 'major-configuration-error'
    context_data = {}

    def __init__(self, hum_event, **kwargs):
        """
        @param hum_event: An Event object. Used to determine the `operational_id`. `None` is accepted value if
                          a suitable Event object is not available.
        """
        self.template = self._get_task_template()
        if hum_event:
            self.context_data.update(hum_event_adapter(hum_event))

        if kwargs:
            self.context_data.update(catch_all_adapter(**kwargs))

    def _get_task_template(self):
        m_path = os.path.join(TASK_TEMPLATES_DIR, '{}.mustache'.format(self._task_template_filename))
        with open(m_path, 'r') as m_file:
            template = m_file.read()

        return template

    def get_task_unique_summary(self):
        return chevron.render(self._primary_key_template, self.context_data, def_ldel='<%', def_rdel='%>')

    def get_task_description(self):
        return chevron.render(self.template, self.context_data, def_ldel='<%', def_rdel='%>')

    def get_operation_id(self):
        """
        @returns The `operational_id` files from the Event object passed to the constructor (if it exists).
                 Else the value `None` is returned.
        """
        try:
            return self.context_data['hum_event']['operation_id']
        except Exception:
            return None


class FixDataNameTask(TaskReferralBase):
    _task_template_filename = 'misnamed-gis-file'
    _primary_key_template = 'gisdata : <%name_result.name_to_validate%> : Incorrectly Named'

    def __init__(self, name_result, cmf):
        super(FixDataNameTask, self).__init__(None)
        self.context_data.update(name_result_adapter(name_result))
        self.context_data.update(cmf_description_adapter(cmf))


class FixFileInWrongDirTask(TaskReferralBase):
    _task_template_filename = 'file-in-wrong-directory'
    _primary_key_template = 'gisdata : ->TBC.folder_name<-'


def cmf_description_adapter(cmf):
    return {'cmf': cmf.__dict__.copy()}


def hum_event_adapter(hum_event):
    return {'hum_event': hum_event.__dict__.copy()}


def catch_all_adapter(**kwargs):
    """
    Used to create generic catch all messages, with unknown context data. Creates a list
    """
    ocl_dict = {'oc_list': [{'other_context': info_str} for info_str in kwargs.items()]}

    return {'catch_all': ocl_dict}


def name_result_adapter(name_result):
    valid_clause_list = []
    invalid_clause_list = []
    if name_result.is_parsable:
        valid_clause_list = [clause.get_message for clause in name_result._asdict().values() if clause.is_valid]
        invalid_clause_list = [clause.get_message for clause in name_result._asdict().values() if not clause.is_valid]

    nr_dict = {
        'name_to_validate': name_result.name_to_validate,
        'is_parsable': name_result.is_parsable,
        'valid_clause_list': [
            {'valid_clause': clause_msg} for clause_msg in sorted(valid_clause_list)],
        'invalid_clause_list': [
            {'invalid_clause': clause_msg} for clause_msg in sorted(invalid_clause_list)]
    }

    return {'name_result': nr_dict}


def layer_adapter(recipe_lyr):
    dict_copy = recipe_lyr.__dict__.copy()
    if dict_copy['data_source_path']:
        dict_copy['data_dir'] = os.path.dirname(dict_copy['data_source_path'])
        dict_copy['data_filename'] = os.path.basename(dict_copy['data_source_path'])
    return {'layer': dict_copy}


def layer_reg_ex_adapter(recipe_lyr, cmf):
    """
    Creates context data to support human readable intepritation of the regex used to
    seach for datasets.
    The key to this adapter is the fact that in most cases the layer's `reg_ex` property
    itself can be parse using the Data Naming Convention.
    """
    nc = name_convention.NamingConvention(cmf.data_nc_definition)
    ncr = nc.validate(recipe_lyr.reg_exp)
    # ncr_context = name_result_adapter(ncr)

    valid_clause_list = []
    invalid_clause_list = []
    if ncr.is_parsable:
        for clause_name, clause_res in ncr._asdict().items():
            clause_value = clause_res._asdict().get('Value', None)
            clause_desc = clause_res._asdict().get('Description', None)
            # TODO: This is hackish. It should really be evident from the DataClauseValues object
            # whether or not it needs handling differently because it a freetext clause. The string
            # matching here is actually matching on something that originates from a config file, not
            # elsewhere within the code.
            if clause_res.is_valid and (clause_name != 'freetext'):
                msg = "The '{}' clause value *must* be '{}' (representing {})".format(
                    clause_name, clause_value, clause_desc)
                valid_clause_list.append(msg)
            else:
                if clause_value:
                    part2 = " that matches this expression '{}'".format(clause_value)
                msg = "The '{}' clause value may be any valid value{}.".format(clause_name, part2)
                invalid_clause_list.append(msg)

    nr_dict = {
        'name_to_validate': ncr.name_to_validate,
        'is_parsable': ncr.is_parsable,
        'valid_clause_list': [
            {'valid_clause': clause_msg} for clause_msg in sorted(valid_clause_list)],
        'invalid_clause_list': [
            {'invalid_clause': clause_msg} for clause_msg in sorted(invalid_clause_list)]
    }

    # Replace the key for the returned dict
    return {'search_info': nr_dict}



from unittest import TestCase
import mapactionpy_controller.check_naming_convention as check_naming_convention
import mapactionpy_controller.name_convention as name_convention
from mapactionpy_controller.crash_move_folder import CrashMoveFolder

import os
import six

# works differently for python 2.7 and python 3.x
if six.PY2:
    import mock  # noqa: F401
    from mock import mock_open, patch
else:
    from unittest import mock  # noqa: F401
    from unittest.mock import mock_open, patch  # noqa: F401


class TestCheckNamingConventionTool(TestCase):

    def setUp(self):
        self.parent_dir = os.path.dirname(
            os.path.dirname(os.path.realpath(__file__)))
        self.cmf_descriptor_path = os.path.join(
            self.parent_dir, 'example', 'cmf_description_flat_test.json')

    def test_get_single_file_checker(self):
        cmf = CrashMoveFolder(self.cmf_descriptor_path)
        nc_desc_path = os.path.join(self.parent_dir, 'example', 'data_naming_convention.json')
        nc = name_convention.NamingConvention(nc_desc_path)

        passing_path = '/path/to/some/gisdata/206_bldg/ken_bldg_bdg_py_s4_osm_pp.shp'
        func = check_naming_convention.get_single_file_checker(passing_path, nc, cmf)
        self.assertIn('parsable and valid', func().get_message)

        failing_path = '/path/to/some/gisdata/202_admn/ken_admn_ad0_ln_s0_IEBC_pp_HDX.shp'
        func = check_naming_convention.get_single_file_checker(failing_path, nc, cmf)
        self.assertRaises(ValueError, func)



import os
import pycountry
from mapactionpy_controller.map_data import MapData
from mapactionpy_controller.map_doc import MapDoc


class XmlExporter:
    def __init__(self, event, chef):
        self.event = event
        self.chef = chef

    def write(self, params):
        # Set up dictionary for all the values required for the export XML file
        exportPropertiesDict = self.setExportParameters(params)
        exportData = MapData(exportPropertiesDict)
        mapDocument = MapDoc(exportData)

        exportXmlFileName = params["coreFileName"]+".xml"
        exportXmlFileLocation = os.path.join(params["exportDirectory"], exportXmlFileName)

        f = open(exportXmlFileLocation, "w")
        f.write(mapDocument.to_xml())
        f.close()

        return exportXmlFileLocation

    def setExportParameters(self, params):
        exportPropertiesDict = {}
        # Copy from params
        exportPropertiesDict["versionNumber"] = params["versionNumber"]
        exportPropertiesDict["mapNumber"] = params["mapNumber"]
        exportPropertiesDict["themes"] = params["themes"]
        exportPropertiesDict["pdffilename"] = params["pdfFileName"]
        exportPropertiesDict["jpgfilename"] = params["jpgFileName"]
        exportPropertiesDict["pdffilesize"] = params["pdfFileSize"]
        exportPropertiesDict["jpgfilesize"] = params["jpgFileSize"]
        exportPropertiesDict["title"] = params["productName"]
        exportPropertiesDict["xmin"] = params["xmin"]
        exportPropertiesDict["ymin"] = params["ymin"]
        exportPropertiesDict["xmax"] = params["xmax"]
        exportPropertiesDict["ymax"] = params["ymax"]
        exportPropertiesDict["ref"] = params["coreFileName"]
        exportPropertiesDict["mxdfilename"] = params["coreFileName"]
        exportPropertiesDict["product-type"] = params["productType"]
        exportPropertiesDict["summary"] = params["summary"]
        # Fixed values
        exportPropertiesDict["imagerydate"] = params.get('imagerydate', "")
        exportPropertiesDict["papersize"] = params.get('papersize', "A3")
        exportPropertiesDict["access"] = params.get('access', "MapAction")
        exportPropertiesDict["accessnotes"] = params.get('accessnotes', "")
        exportPropertiesDict["location"] = params.get('location', "")
        exportPropertiesDict["qclevel"] = params.get('qclevel', "Automatically generated")
        exportPropertiesDict["qcname"] = params.get('qcname', "")
        exportPropertiesDict["proj"] = params.get('proj', "")
        exportPropertiesDict["datasource"] = params.get('datasource', "")
        exportPropertiesDict["kmlresolutiondpi"] = params.get('kmlresolutiondpi', "")
        exportPropertiesDict["paperxmax"] = params.get('paperxmax', "")
        exportPropertiesDict["paperxmin"] = params.get('paperxmin', "")
        exportPropertiesDict["paperymax"] = params.get('paperymax', "")
        exportPropertiesDict["paperymin"] = params.get('paperymin', "")
        exportPropertiesDict["createdate"] = params.get('createdate', "")
        exportPropertiesDict["createtime"] = params.get('createtime', "")
        exportPropertiesDict["scale"] = params.get('scale', "")
        exportPropertiesDict["datum"] = params.get('datum', "")
        exportPropertiesDict["language-iso2"] = self.event.language_iso2
        exportPropertiesDict["pdfresolutiondpi"] = self.event.default_pdf_res_dpi
        exportPropertiesDict["jpgresolutiondpi"] = self.event.default_jpeg_res_dpi
        exportPropertiesDict["countries"] = self.event.country_name
        exportPropertiesDict["glideno"] = self.event.glide_number
        exportPropertiesDict["operationID"] = self.event.operation_id
        exportPropertiesDict["sourceorg"] = self.event.default_source_organisation

        if (params["versionNumber"] == 1):
            exportPropertiesDict["status"] = "New"
        else:
            exportPropertiesDict["status"] = "Update"

        language = pycountry.languages.get(alpha_2=self.event.language_iso2)
        if (language is not None):
            exportPropertiesDict["language"] = language.name
        else:
            exportPropertiesDict["language"] = None

        if (self.chef is not None):
            exportPropertiesDict["createdate"] = self.chef.createDate
            exportPropertiesDict["createtime"] = self.chef.createTime
            exportPropertiesDict["scale"] = self.chef.scale()
            exportPropertiesDict["datum"] = self.chef.spatialReference()

        # return
        return exportPropertiesDict



from mapactionpy_controller.plugin_base import BaseRunnerPlugin
from mapactionpy_controller.event import Event
from mapactionpy_controller.crash_move_folder import CrashMoveFolder
from mapactionpy_controller.layer_properties import LayerProperties
from mapactionpy_controller.map_recipe import MapRecipe
import fixtures
import os
from unittest import TestCase, skip
import six
import sys

# works differently for python 2.7 and python 3.x
if six.PY2:
    import mock  # noqa: F401
else:
    from unittest import mock  # noqa: F401


class DummyRunner(BaseRunnerPlugin):
    def __init__(self, hum_event):
        super(DummyRunner, self).__init__(hum_event)

    def get_aspect_ratios_of_templates(self, possible_templates, recipe):
        return [
            ('one', 1.0),
            ('two', 2.0)
        ]

    def export_maps(self, **kwargs):
        return kwargs['state']

    def build_project_files(self, **kwargs):
        return kwargs['state']

    def get_projectfile_extension(self):
        return '.dummy_project_file'

    def get_lyr_render_extension(self):
        return '.lyr'

    def create_ouput_map_project(self, **kwargs):
        return kwargs['state']


class TestPluginBase(TestCase):
    def setUp(self):
        self.parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.dir_to_valid_cmf_des = os.path.join(self.parent_dir, 'example')
        self.path_to_valid_cmf_des = os.path.join(self.dir_to_valid_cmf_des, 'cmf_description_flat_test.json')
        self.path_to_event_des = os.path.join(self.dir_to_valid_cmf_des, 'event_description.json')
        self.cmf = CrashMoveFolder(self.path_to_valid_cmf_des)
        self.lyr_props = LayerProperties(self.cmf, '', verify_on_creation=False)
        self.dummy_runner = DummyRunner(Event(self.path_to_event_des))

    def test_get_all_templates_by_regex(self):
        recipe = mock.Mock(name='mock_recipe')
        recipe.template = 'abcde'

        dummy_map_templates = '/xyz/'

        if sys.platform == 'win32':
            dummy_map_templates = 'C:\\xyz\\'

        self.dummy_runner.cmf.map_templates = dummy_map_templates

        available_templates = [
            'one-two-three.dummy_project_file',
            'one-two-three.txt',
            'abcde.dummy_project_file',
            'abcde.txt'
        ]

        expect_result = [
            '{}abcde.dummy_project_file'.format(dummy_map_templates)
        ]

        with mock.patch('mapactionpy_controller.plugin_base.os.listdir') as mock_listdir:
            with mock.patch('mapactionpy_controller.plugin_base.os.path.isfile') as mock_isfile:
                mock_listdir.return_value = available_templates
                mock_isfile.return_value = True

                actual_result = self.dummy_runner._get_all_templates_by_regex(recipe)

        self.assertEqual(actual_result, expect_result)

    def test_get_template_by_aspect_ratio(self):
        template_aspect_ratios = [
            ('one',   1.1),
            ('two',   2),
            ('three', 2.32),
            ('four',  3.1415),
            ('five',  5),
            ('six',   10)
        ]

        # half way between 5.0 and 10.0 ==> 10^(0.5*(log(10)+LOG(5)) = 7.071067812
        test_aspect_ratios = [
            (5.0, 'five'),
            (4.9, 'five'),
            (5.1, 'five'),
            (100, 'six'),
            (0.0001, 'one'),
            (7.0711, 'six'),
            (7.0710, 'five')
        ]

        for target_ar, expect_result in test_aspect_ratios:
            actual_result = self.dummy_runner._get_template_by_aspect_ratio(template_aspect_ratios, target_ar)
            # print('expect_result={}, actual_result={}'.format(expect_result, actual_result))
            self.assertEqual(expect_result, actual_result)

        template_aspect_ratios = [
            ('landscape_bottom',	1.975),
            ('landscape_side',		1.294117647),
            ('portrait'	,	0.816816817)
        ]

        # linear_aspect_ratios = [
        #     (1.623, 'landscape_side'),
        #     (1.036, 'portrait'),
        #     (1.038, 'portrait'),
        #     (1.031, 'portrait')
        # ]

        log_aspect_ratios = [
            (1.623, 'landscape_bottom'),
            (1.036, 'landscape_side'),
            (1.038, 'landscape_side'),
            (1.031, 'landscape_side')
        ]

        for target_ar, expect_result in log_aspect_ratios:
            actual_result = self.dummy_runner._get_template_by_aspect_ratio(template_aspect_ratios, target_ar)
            # print('expect_result={}, actual_result={}'.format(expect_result, actual_result))
            self.assertEqual(expect_result, actual_result)

    def test_get_templates(self):

        # Case 1
        # Pre-existing valid `map_project_path` value
        expect_result = MapRecipe(fixtures.recipe_test_for_search_for_shapefiles, self.lyr_props)
        test_recipe = MapRecipe(fixtures.recipe_test_for_search_for_shapefiles, self.lyr_props)
        expect_result.map_project_path = '/path/that/exists.mxd'
        test_recipe.map_project_path = '/path/that/exists.mxd'

        with mock.patch('mapactionpy_controller.plugin_base.os.path.exists') as mock_path_exists:
            mock_path_exists.return_value = True
            actual_result = self.dummy_runner.get_templates(state=test_recipe)

        self.assertEqual(actual_result, expect_result)

        # Case 2
        # Non-None `map_project_path` value which points of non-existing file
        test_recipe = MapRecipe(fixtures.recipe_test_for_search_for_shapefiles, self.lyr_props)
        test_recipe.map_project_path = '/path/that/does/not/exists.mxd'

        with mock.patch('mapactionpy_controller.plugin_base.os.path.exists') as mock_path_exists:
            mock_path_exists.return_value = False
            with self.assertRaises(ValueError):
                self.dummy_runner.get_templates(state=test_recipe)

        # Case 3 & 4
        # mf.extent exists and has a valid value
        # mf.entent is None
        test_extents = [
            ('one', (1, 1, 5, 5)),
            ('two', (1, 1, 9, 5)),
            ('one', None)
        ]

        for expected_result, extent in test_extents:
            print(expected_result, extent)
            test_recipe = MapRecipe(fixtures.recipe_test_for_search_for_shapefiles, self.lyr_props)
            test_recipe.map_frames[0].extent = extent
            # mf.extent = extent
            actual_recipe = self.dummy_runner.get_templates(state=test_recipe)
            self.assertEquals(expected_result, actual_recipe.template_path)

    @skip('Not ready yet')
    def test_get_next_map_version_number(self):
        self.fail()

    @skip('Not ready yet')
    def test_create_ouput_map_project(self):
        self.fail()

    @skip('Not ready yet')
    def test_export_maps(self):
        self.fail()

    @skip('Not ready yet')
    def test_create_export_dir(self):
        self.fail()

    @skip('Not ready yet')
    def test_zip_exported_files(self):
        self.fail()



import json
import os
from mapactionpy_controller.recipe_layer import RecipeLayer
from mapactionpy_controller.crash_move_folder import CrashMoveFolder


class LayerProperties:
    """
    LayerProperties

    It is expected that a seperate LayerProperties object will be created for each layer_rendering
    file type.
    """

    def __init__(self, cmf, extension, verify_on_creation=True):
        """
        Positional Arguments:
            * cmf: Either a CrashMoveFolder object or a path to a cmf_description.json file. If it
                   is a CrashMoveFolder object and cmf.verify_paths() returns False then an
                   ValueError exception will be rasied.
            * extension: The file extension of the layer_rendering file type. (typically `.lyr` for
                         ESRI Layer files, `'.qml` for QGIS style)

        Optional Named Arguments
            verify_on_creation: True by default. If True then
            `verify_match_with_layer_rendering_dir()` will be called from the constructor.

        """
        try:
            if cmf.verify_paths():
                self.cmf = cmf
            else:
                raise ValueError('The `cmf` parameter for LayerProperties.__init__() can only accept'
                                 'values where the paths verify. eg `cmf.verify_paths() == True`.'
                                 'The value passed in this case failed this test')
        except AttributeError:
            self.cmf = CrashMoveFolder(cmf, verify_on_creation=True)

        if len(extension) == 0 or extension.startswith('.'):
            self.extension = extension
        else:
            self.extension = '.{}'.format(extension)

        self.properties = {}  # Dictionary
        self._parse()

        if verify_on_creation:
            msg = self.get_difference_with_other_layer_set(
                self._get_lyr_rendering_names_as_set(),
                self._get_mismatch_with_layer_rendering_message
            )
            if msg:
                raise ValueError(msg)

    def _parse(self):
        """
        Reads layer properties file
        """
        with open(self.cmf.layer_properties) as json_file:
            jsonContents = json.load(json_file)
            for layer in jsonContents['layerProperties']:
                mapLayer = RecipeLayer(layer, self)
                self.properties[mapLayer.name] = mapLayer

    def _get_lyr_rendering_names_as_set(self):
        files_unique = set()
        dir_content = os.listdir(self.cmf.layer_rendering)
        for f in dir_content:
            f_path = os.path.join(self.cmf.layer_rendering, f)
            filename, fileext = os.path.splitext(f)
            if (os.path.isfile(f_path)) and (fileext == self.extension):
                files_unique.add(filename)

        return files_unique

    def is_difference_with_layer_rendering_dir(self):
        msg = self.get_difference_with_other_layer_set(
            self._get_lyr_rendering_names_as_set(),
            self._get_mismatch_with_layer_rendering_message
        )
        return bool(msg)

    def get_difference_with_other_layer_set(self, other_lyrs, get_msg_func):
        """
        Ensures that there is a one-to-one correspondance between
        * The layers described by this LayerProperties Object
        * Another set() listing layers from any source.

        For example this method is used to verify the consistency between
        * The layer_properties.json file and the files listed in the `cmf.layer_rendering` directory
        * The layer_properties.json file and the map_cookbook.json file.

        :returns:
            * `None`: if there are no differences between the LayerProperties and the `other_lyrs` set.
            * A message describing the differences between the two. The message is constructed by
              the `get_msg_func` supplied. The supplied funtion must accept two sets `lp_only` and `others_only`.
              Each of the two sets incldue onyl those layers which occur uniquely in each. The message is
              constructed like so:
                  msg = get_msg_func(lp_only, others_only)
        """
        lp_unique_lyrs = set(self.properties)

        lp_only = lp_unique_lyrs.difference(other_lyrs)
        others_only = other_lyrs.difference(lp_unique_lyrs)

        msg = None
        if len(lp_only) or len(others_only):
            msg = get_msg_func(lp_only, others_only)

        return msg

    def _get_mismatch_with_layer_rendering_message(self, lp_only, files_only):
        msg = ('There is a mismatch between:\n'
               ' (a) The layers described in the layer_properties.json file "{}"\n'
               ' and (b) The layers files, with file extension "{}", listed in the'
               ' `cmf.layer_rendering` directory "{}".\n'
               ' Either ensure that these files match, or create the LayerProperties object using the'
               ' parameter `verify_on_creation=False`\n'.format(
                   self.cmf.layer_properties,
                   self.extension,
                   self.cmf.layer_rendering
               ))

        pair = ((lp_only, "The following layers are only present in layer properties json file:"),
                (files_only, "The following files are only present in layer rendering directory:"))

        return self._msg_builder(pair, msg)

    def _msg_builder(self, pair, msg):
        for lyrs, s in pair:
            if len(lyrs):
                msg = msg + '\n{}\n\t'.format(s)
                msg = msg + '\n\t'.join(lyrs)

        return msg



import logging
import os
# import shapefile
import fiona
import glob
import re
import hashlib
import geopandas

import jsonschema

import mapactionpy_controller.data_schemas as data_schemas
import mapactionpy_controller.state_serialization as state_serialization
from mapactionpy_controller import _get_validator_for_config_schema
import mapactionpy_controller.task_renderer as task_renderer


logger = logging.getLogger(__name__)
validate_against_layer_schema = _get_validator_for_config_schema('layer_properties-v0.2.schema')


class FixMissingGISDataTask(task_renderer.TaskReferralBase):
    _task_template_filename = 'missing-gis-file'
    _primary_key_template = 'Could not find data for <%layer.name%>'

    def __init__(self, hum_event, recipe_lyr, cmf):
        super(FixMissingGISDataTask, self).__init__(hum_event)
        self.context_data.update(task_renderer.layer_adapter(recipe_lyr))
        self.context_data.update(task_renderer.cmf_description_adapter(cmf))
        self.context_data.update(task_renderer.layer_reg_ex_adapter(recipe_lyr, cmf))


class FixMultipleMatchingFilesTask(task_renderer.TaskReferralBase):
    _task_template_filename = 'multiple-matching-files'
    _primary_key_template = 'More than one dataset available for <%layer.name%>'

    def __init__(self, hum_event, recipe_lyr, cmf, datasources_list):
        super(FixMultipleMatchingFilesTask, self).__init__(hum_event)
        self.context_data.update(task_renderer.layer_adapter(recipe_lyr))
        self.context_data.update(task_renderer.cmf_description_adapter(cmf))
        self.context_data.update(task_renderer.layer_reg_ex_adapter(recipe_lyr, cmf))
        # Roll-our-own one-line adapter here:
        self.context_data.update({
            'datasources_list': [{'datasources': datasources} for datasources in sorted(datasources_list)]
        })
        target_dir = os.path.dirname(datasources_list.pop())
        self.context_data.update({
            'datasources_dir': target_dir
        })


class FixSchemaErrorTask(task_renderer.TaskReferralBase):
    _task_template_filename = 'schema-error'
    _primary_key_template = 'Schema error in dataset <%layer.data_filename%>'

    def __init__(self, hum_event, recipe_lyr, validation_error, instance_list):
        super(FixSchemaErrorTask, self).__init__(hum_event)
        self.context_data.update(task_renderer.layer_adapter(recipe_lyr))
        self.validation_error = validation_error
        # See https://trello.com/c/dcC8JpYf/180-implenment-task-descriptions-for-all-gis-data-related-tasks
        # self.context_data.update(do_something_with(validation_error))
        i_set = set(instance_list.keys())
        self.context_data.update({
            'data_fields': [{'extant_fields': r_field} for r_field in sorted(i_set)]
        })
        if 'required' in recipe_lyr.data_schema:
            r_set = set(recipe_lyr.data_schema['required'])
            diff_fields = r_set.difference(i_set)

            self.context_data.update({
                'data_schema': [{'required_fields': r_field} for r_field in sorted(r_set)]
            })
            self.context_data.update({
                'missing_fields': [{'fields': r_field} for r_field in sorted(diff_fields)]
            })


class LabelClass:
    """
    Enables selection of properties to support labels in a Layer
    """

    def __init__(self, row):
        self.class_name = row["class_name"]
        self.expression = row["expression"]
        self.sql_query = row["sql_query"]
        self.show_class_labels = row["show_class_labels"]

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Overrides the default implementation (unnecessary in Python 3)"""
        return not self.__eq__(other)


class RecipeLayer:

    OPTIONAL_FIELDS = ('crs', 'data_name', 'data_schema', 'data_source_checksum', 'data_source_path',
                       'error_messages', 'extent', 'layer_file_checksum', 'success',
                       'use_for_frame_extent', 'visible')

    def __init__(self, layer_def, lyr_props, verify_on_creation=True):
        """Constructor.  Creates an instance of layer properties

        Arguments:
            row {dict} -- From the layerProperties.json file
        """
        validate_against_layer_schema(layer_def)

        # Required fields
        self.name = layer_def["name"]
        self.reg_exp = layer_def["reg_exp"]
        self.definition_query = layer_def["definition_query"]
        self.schema_definition = layer_def["schema_definition"]

        self.display = layer_def["display"]
        self.add_to_legend = layer_def["add_to_legend"]
        self.label_classes = list()
        for lbl_class_def in layer_def["label_classes"]:
            self.label_classes.append(LabelClass(lbl_class_def))

        self._get_layer_file_path(layer_def, lyr_props, verify_on_creation)

        # Optional fields
        self._get_data_schema(layer_def, lyr_props)
        self.layer_file_checksum = layer_def.get('layer_file_checksum', self._calc_layer_file_checksum())
        self.data_source_path = layer_def.get('data_source_path', None)
        if self.data_source_path:
            self.data_source_path = os.path.abspath(self.data_source_path)

        self.data_source_checksum = layer_def.get('data_source_checksum', self._calc_data_source_checksum())
        self.data_name = layer_def.get('data_name', None)
        self.extent = layer_def.get('extent', None)
        self.crs = layer_def.get('crs', None)
        self.error_messages = layer_def.get('error_messages', [])
        self.success = layer_def.get('success', False)
        self.visible = layer_def.get('visible', True)
        self._apply_use_for_frame_extent(layer_def)

    def _apply_use_for_frame_extent(self, layer_def):
        # Allow for three valid values True/False/None
        if 'use_for_frame_extent' in layer_def:
            self.use_for_frame_extent = bool(layer_def['use_for_frame_extent'])
        else:
            self.use_for_frame_extent = None

    def _get_layer_file_path(self, layer_def, lyr_props, verify_on_creation):
        if 'layer_file_path' in layer_def:
            self.layer_file_path = os.path.abspath(layer_def['layer_file_path'])
            if verify_on_creation:
                self.verify_layer_file_path()

        else:
            self.layer_file_path = os.path.abspath(os.path.join(
                lyr_props.cmf.layer_rendering,
                (self.name + lyr_props.extension)
            ))

    def _get_data_schema(self, layer_def, lyr_props):
        if 'data_schema' in layer_def:
            self.data_schema = layer_def['data_schema']
        else:
            schema_file = os.path.abspath(os.path.join(lyr_props.cmf.data_schemas, self.schema_definition))
            self.data_schema = data_schemas.parse_yaml(schema_file)

        # check that the schema itself is valid.
        jsonschema.Draft7Validator.check_schema(self.data_schema)

    def verify_layer_file_path(self):
        if not os.path.exists(self.layer_file_path):
            raise ValueError("The expected layer file {} could not be found."
                             "".format(self.layer_file_path))

    def _check_lyr_is_in_recipe(self, recipe):
        # This is being paraniod. This can only occur if there are more than one MapRecipe objects
        # being proceed simulatiously. There is isn't currently a use case where that would occur in
        # production code. This check is present just in case.
        if self not in recipe.all_layers():
            error_msg = 'Attempting to update a layer ("{}") which is not part of the recipe'.format(
                self.name)
            logging.error(error_msg)
            raise ValueError(error_msg)

    def get_data_finder(self, cmf, all_gis_files):
        """
        This method returns a function which tests for the existance of data that matches
        the param `recipe_lyr.reg_exp`.

        Create a new function for each layer within a recipe.
        """
        # Get list of files, so that they are only queried on disk once.
        # Make this into a list of full_paths (as returned by `get_all_gisfiles(cmf)`) and
        # just the file name
        # all_gis_files = [(f_path, os.path.basename(f_path)) for f_path in get_all_gisfiles(cmf)]

        def _data_finder(**kwargs):
            recipe = kwargs['state']

            self._check_lyr_is_in_recipe(recipe)

            found_files = []

            # Match filename *including extension* against regex
            # But only store the filename without extension
            found_files.extend(
                [(f_path, os.path.splitext(f_name)[0])
                 for f_path, f_name in all_gis_files if re.search(self.reg_exp, f_name, re.IGNORECASE)]
            )

            # Do checks and raise exceptions if required.
            self._check_found_files(found_files, cmf, recipe.hum_event)

            # else assume everthing is OK:
            self.data_source_path, self.data_name = found_files.pop()
            self.data_source_checksum = self._calc_data_source_checksum()

            return recipe

        return _data_finder

    def _check_found_files(self, found_files, cmf, hum_event):
        """
        This method checks the list of files in `get_lyr_data_finder`. It is called within `get_lyr_data_finder`
        so there is no need to call it seperately. A 'ValueError' exception, along with a Task and task
        description will be raised if there are too few or too many files in the `found_files` list.

        @param found_files: A list of tuples. Each tuple should be represent one of the files found and the tuple
                            should be in the format `(full_path, filename_without_extension)`
        @param recipe_lyr: The recipe being processed. Used for providing contact to any error message that may
                        be created.
        @param cmf: The crash move folder being searched. Used for providing contact to any error message that
                    may be created.
        @raises ValueError: If there is not exactly one file in the found_files lilst.
        """
        # If no data matching is found:
        # Test on list of paths as they are guarenteed to be unique, whereas base filenames are not
        if not found_files:
            self.error_messages.append('Unable to find dataset for this layer')
            missing_data_task = FixMissingGISDataTask(hum_event, self, cmf)
            raise ValueError(missing_data_task)

        # If multiple matching files are found
        if len(found_files) > 1:
            self.error_messages.append('Found multiple datasets which match this layer')
            found_datasources = [f_path for f_path, f_name in found_files]
            multiple_files_task = FixMultipleMatchingFilesTask(hum_event, self, cmf, found_datasources)
            raise ValueError(multiple_files_task)

    def _calc_data_source_checksum(self):

        def files_in_shp_file():
            base_path = os.path.splitext(self.data_source_path)[0]
            files = glob.glob('{}*'.format(base_path))
            return [f_path for f_path in files if (os.path.isfile(f_path) and not f_path.endswith('.lock'))]

        def files_in_dir():
            return [os.path.join(f_path, f_name) for f_path, d_name, f_name in os.walk(self.data_source_path)]

        f_list = []

        if self.data_source_path and (os.path.isfile(self.data_source_path)):
            f_list = files_in_shp_file()

        if self.data_source_path and (os.path.isdir(self.data_source_path)):
            f_list = files_in_dir()

        f_list.sort()

        hash = hashlib.md5()
        for f_path in f_list:
            with open(f_path, "rb") as data_file:
                hash.update(data_file.read())

        return hash.hexdigest()

    def _calc_layer_file_checksum(self):
        # if not os.path.isfile(self.layer_file_path):
        #     return None

        hash = hashlib.md5()
        if os.path.isfile(self.layer_file_path):
            with open(self.layer_file_path, "rb") as lf:
                hash.update(lf.read())
        return hash.hexdigest()

    # def get_schema_checker(self, runner):
    def check_data_against_schema(self, **kwargs):
        logging.debug('Attempting to validate jsonschema for gis data')
        if not self.data_source_path:
            error_msg = ('Cannot check data schema for layer ("{}") before the relevant data has been found.'
                         ' Please use `get_data_finder()` first.'.format(self.name))
            logging.debug(error_msg)
            raise ValueError(error_msg)

        recipe = kwargs['state']

        if not self.data_source_path.endswith('.shp'):
            logging.info("Unable to check schema on for data sources which aren't shapefiles."
                         " Skipping schema check on file: {}".format(self.data_source_path))
            return recipe

        # Validate
        try:
            # Check for self consistancy before proceeding
            self._check_lyr_is_in_recipe(recipe)

            # Only load one row since we are only checking the schema
            gdf = geopandas.read_file(self.data_source_path, rows=1)
            # Make columns needed for validation
            gdf['geometry_type'] = gdf['geometry'].apply(lambda x: x.geom_type)
            # print(gdf.crs)
            if isinstance(gdf.crs, dict):
                gdf['crs'] = gdf.crs['init']
            else:
                gdf['crs'] = gdf.crs

            instance_list = gdf.to_dict('list')
            jsonschema.validate(instance=instance_list, schema=self.data_schema)
            logging.debug('Successfully validates jsonschema for gis data')
        # except jsonschema.ValidationError as jsve:
        except Exception as exp:
            error_msg = ('Failed whilst check data schema for layer ("{}") with exception: {}'.format(
                self.name, exp))
            logging.debug(error_msg)
            self.error_messages.append('Data schema check failed')
            fset = FixSchemaErrorTask(recipe.hum_event, self, exp, instance_list)
            raise ValueError(fset)

        return recipe

    def calc_extent(self, **kwargs):
        """
        Extracts the Coordinate Reference System (crs) and extent for the dataset located at
        `data_source_path`. This method will update `self.crs` and `self.extent`.

        @raises ValueError: If the `self.data_source_path` has not be set or is invalid.
        """
        if not self.data_source_path:
            logger.info('Have no self.data_source_path')
            raise ValueError(
                'Cannot calculate bounding box until relevant data has been found.'
                ' Please use `get_data_finder()` first.')

        logger.info('Has self.data_source_path')
        recipe = kwargs['state']
        # print('recipe before in layer.calc_extent:')
        # print(recipe)
        self._check_lyr_is_in_recipe(recipe)

        logger.info('passed _check_lyr_is_in_recipe')
        sf = fiona.open(self.data_source_path)
        self.extent = sf.bounds
        self.crs = sf.crs['init']

        return recipe

    def __eq__(self, other):
        try:
            return self.__dict__ == other.__dict__
        except AttributeError:
            return False

    def __ne__(self, other):
        """Overrides the default implementation (unnecessary in Python 3)"""
        return not self.__eq__(other)

    def __getstate__(self):
        return state_serialization.get_state_optional_fields(self, RecipeLayer.OPTIONAL_FIELDS)

    def __setstate__(self, state):
        state_serialization.set_state_optional_fields(self, state, RecipeLayer.OPTIONAL_FIELDS)



import json
import logging
from os import path

from jsonschema import validate

CONFIG_SCHEMAS_DIR = path.join(path.abspath(path.dirname(__file__)), 'schemas')
TASK_TEMPLATES_DIR = path.join(path.abspath(path.dirname(__file__)), 'task-templates')


def _get_validator_for_config_schema(schema_file):

    schema_path = path.join(CONFIG_SCHEMAS_DIR, schema_file)
    with open(schema_path) as sf:
        schema = json.load(sf)

    def validate_against_schema(data):
        validate(data, schema)

    return validate_against_schema


# logging.basicConfig(
#     level=logging.DEBUG,
#     format=(
#         '%(asctime)s %(module)s %(name)s.%(funcName)s +%(lineno)s: %(levelname)-8s'
#         ' [%(process)d] %(message)s',
#     )
# )

logger = logging.getLogger(__name__)
# logger = logging.getLogger()
logger.setLevel(logging.INFO)
# create file handler which logs even debug messages
# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
# create formatter and add it to the handlers
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s (%(module)s +ln%(lineno)s) ;- %(message)s')
# formatter = logging.Formatter('%(asctime)s %(module)s %(name)s.%(funcName)s
# +%(lineno)s: %(levelname)-8s [%(process)d] %(message)s')

ch.setFormatter(formatter)
# add the handlers to the logger
logger.addHandler(ch)



from unittest import TestCase, skip
import mapactionpy_controller.cli as cli
import sys
import os

# works differently for python 2.7 and python 3.x
try:
    from unittest import mock
except ImportError:
    import mock


class TestCLI(TestCase):

    def setUp(self):
        self.parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.path_to_cmf_file = os.path.join(self.parent_dir, 'example', 'cmf_description_flat_test.json')
        self.path_to_event_file = os.path.join(self.parent_dir, 'example', 'event_description.json')
        self.nonexistant_path = '/file/that/does/not/exist'

    @mock.patch('mapactionpy_controller.main_stack.hft.connected_to_terminal')
    @mock.patch('mapactionpy_controller.check_naming_convention.glob.glob')
    @mock.patch('mapactionpy_controller.cli.process_stack')
    def test_cli_with_steps(self, mock_stack, mock_glob, mock_hft):
        # mock_hft is purely to prevent the console from becoming too clutered when running unittests
        mock_hft.return_value = False

        # valid CLI options that will call a list of steps
        input_args_list = [
            ['defaultcmf', '--verify', self.path_to_cmf_file],
            ['defaultcmf', '--verify', self.path_to_cmf_file],
            ['gisdata', '--verify', self.path_to_event_file],
            ['maps', '--build', self.path_to_event_file]
        ]

        # Note that becuase `process_stack` can be called more than once per CLI invokation
        # it is possible that `input_msg_list` may be a different length to `input_args_list`
        input_msg_list = [
            'map template naming convention',
            'layer properties json file and the MapCookbook',
            'data naming convention',
            'Humanitarian Event description file',
            'MapCookbook files'
        ]

        glob_list = [
            '/path/to/some/gisdata/202_admn/ken_admn_ad0_ln_s0_IEBC_pp_HDX.shp',
            '/path/to/some/gisdata/202_admn/ken_admn_ad0_py_s1_IEBC_pp_HDX.shp',
            '/path/to/some/gisdata/202_admn/ken_admn_ad4_py_s1_HDX_pp_CBS.shp',
            '/path/to/some/gisdata/206_bldg/ken_bldg_bdg_py_s4_osm_pp.shp',
            '/path/to/some/gisdata/209_cccm/eafr_cccm_ref_pt_s1_unhcr_pp_CampLocations.shp',
            '/path/to/some/gisdata/211_elev/ken_elev_cst_ln_s0_iebc_pp_HDX.shp'
        ]

        mock_glob.return_value = glob_list

        # with mock.patch('mapactionpy_controller.cli.process_stack') as mock_steps:
        for test_args in input_args_list:
            # set the commandline and run the production code
            sys.argv[1:] = test_args
            cli.entry_point()

        # Now get the information out of the mock
        all_calls = mock_stack.call_args_list

        for call, input_msg in zip(all_calls, input_msg_list):
            # `call` is a tuple.
            # `call[0]` is the list the list of _positional_ args to mock_steps
            # `call[1]` is the list of the _keyword_ args to mock_steps
            # therefore `call[0][0]` is the first _positional_ arg, which is itself a
            # list of Step objects
            # Just aggreegate all of the running msg from all of the steps together
            out_msg = "".join(stp.running_msg for stp in call[0][0])
            # check that our test string occurs somewhere within this string
            self.assertIn(input_msg, out_msg)

    def test_cli_without_steps(self):

        # valid CLI options that will be handled with the 'cli' module
        not_implenmented_combos = [
            ['defaultcmf', self.path_to_cmf_file],
            ['humevent', self.path_to_event_file],
            ['gisdata', self.path_to_event_file],
            ['maps', self.path_to_event_file]
        ]

        for testargs in not_implenmented_combos:
            # set the commandline and run the production code
            sys.argv[1:] = testargs
            self.assertRaises(NotImplementedError, cli.entry_point)

        invalid_combos = [
            [''],
            ['defaultcmf'],
            ['defaultcmf', '--verify'],
            ['defaultcmf', '/path/to/non/exsistant/file'],
            ['gisdata'],
            ['gisdata', '--verify'],
            ['non-exitant-sub-cmd']
        ]

        for testargs in invalid_combos:
            # set the commandline and run the production code
            sys.argv[1:] = testargs
            self.assertRaises(SystemExit, cli.entry_point)

    @skip('Not implenmented yet')
    def test_cli_with_cmf_deduced_from_current_dir(self):
        # TODO
        self.fail()



import os
from unittest import TestCase
import mapactionpy_controller.plugin_controller as plugin_controller
import mapactionpy_controller.main_stack as main_stack
from mapactionpy_controller.tests.test_plugin_base import DummyRunner
from mapactionpy_controller.steps import Step
from mapactionpy_controller.crash_move_folder import CrashMoveFolder
from mapactionpy_controller.event import Event
from mapactionpy_controller.map_cookbook import MapCookbook
from mapactionpy_controller.layer_properties import LayerProperties
# works differently for python 2.7 and python 3.x
try:
    from unittest import mock
except ImportError:
    import mock


class TestPluginController(TestCase):

    def setUp(self):
        self.parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.path_to_cmf_file = os.path.join(self.parent_dir, 'example', 'cmf_description_flat_test.json')
        # print('self.path_to_cmf_file = {}'.format(self.path_to_cmf_file))
        self.path_to_event_file = os.path.join(self.parent_dir, 'example', 'event_description.json')
        # self.nonexistant_path = '/file/that/does/not/exist'

    @mock.patch('mapactionpy_controller.main_stack.parse_feedback')
    def test_get_cookbook_steps(self, mock_pf):
        # This asserts that a list of steps is added to the stack when `get_cookbook_steps` if called
        # First do the work:
        test_runner = DummyRunner(Event(self.path_to_event_file))
        test_runner.cmf = CrashMoveFolder(self.path_to_cmf_file)
        initial_step = plugin_controller.get_cookbook_steps(test_runner, None, dry_run=True, verify_on_creation=False)
        main_stack.process_stack(initial_step, None)

        # Now get the information out of the mock:
        found_list_of_steps = False
        for call in mock_pf.call_args_list:
            try:
                # print()
                # print('---------------------')
                # print('test_get_cookbook_steps')
                # print(call)
                # print('END:test_get_cookbook_steps')
                # print('---------------------')
                # retrieved_result = call[1]['result']
                retrieved_result = call[1].get('result', None)
                if all([isinstance(stp, Step) for stp in retrieved_result]):
                    # print('found_list_of_steps=True')
                    found_list_of_steps = True
            except TypeError:
                pass

        self.assertTrue(found_list_of_steps)

    def test_select_recipes(self):
        # This looks for the MapID in the step.running_msg for the selected maps

        # A list of tuples to test
        # * The param (as entered by a human) to be passed to `get_cookbook_steps`. Includes upper, lower
        #   and mixed case, MapIDs that aren't in the cookbook etc. Can be either a string or a list
        # * A list of the mapIDs that should be created (exactly as in the cookbook file - match case etc)
        # * A list of mapID that should NOT be created (and represent a failure)
        #
        # This is the list of mapID that exist in the relevant cookbook
        #  ['MA001', 'MA002', 'MA003', 'MA004']
        test_cases = [
            ('ma001', ['MA001'], ['MA002', 'MA003', 'MA004']),
            ('MA004', ['MA004'], ['MA001', 'MA002', 'MA003']),
            ('Ma004', ['MA004'], ['MA001', 'MA002', 'MA003']),
            ('MA001', ['MA001'], ['MA002', 'MA003', 'MA004']),
            (['mA001', 'ma002'], ['MA001', 'MA002'], ['MA003', 'MA004']),
            ('ma999', [], ['MA001', 'MA002', 'MA003', 'MA004']),
            (None, ['MA001', 'MA002', 'MA003', 'MA004'], [])
        ]

        cmf = CrashMoveFolder(self.path_to_cmf_file)
        lp = LayerProperties(cmf, 'test', False)
        cb = MapCookbook(cmf, lp, False)

        for mapid_arg, should_create, fail_list in test_cases:
            # print('mapid_arg = {}'.format(mapid_arg))
            # print('should_create = {}'.format(should_create))
            # print('fail_list = {}'.format(fail_list))
            should_create_set = set(should_create)

            selected_recipes = plugin_controller.select_recipes(cb, mapid_arg)
            selected_map_ids = set([recipe.mapnumber.upper() for recipe in selected_recipes])

            # Check that no MapID were selected that shouldn't have been:
            self.assertTrue(selected_map_ids.isdisjoint(fail_list))

            # Check that all of the `selected_ids` match the `should_create` set:
            self.assertEqual(should_create_set, selected_map_ids)



import json
import re
from collections import namedtuple
from pydoc import locate

from mapactionpy_controller.name_clause_validators import NamingClause


class NamingConvention:
    def __init__(self, nc_json_path):
        self.nc_json_path = nc_json_path
        self._clause_validation = {}

        with open(self.nc_json_path) as json_file:
            json_contents = json.load(json_file)

        self.regex = re.compile(json_contents['pattern'])

        rx_grp_list = self.regex.groupindex.keys()

        json_clause_names = set()
        for clause_def in json_contents['clauses']:
            json_clause_names.add(clause_def['name'])

        if not (set(rx_grp_list) == json_clause_names):
            raise NamingException(
                'Error in {}. Mismatch between clause definition {} '
                'and groups name in regular expresion {}')

        for clause_def in json_contents['clauses']:
            clause_name = clause_def['name']
            validator_name = clause_def['validator']

            try:
                Validator = locate(validator_name)
                # Instantiate the class (pass arguments to the constructor, if needed)
                dnlc = Validator(self.nc_json_path, **clause_def)
            except TypeError:
                raise NamingException('Error in {}. The validation type {} cannot be loaded'
                                      ''.format(self.nc_json_path, validator_name))

            if isinstance(dnlc, NamingClause):
                self._clause_validation[clause_name] = dnlc
            else:
                raise NamingException('Error in {}. The specified validator class {} is not '
                                      'an instance of mapactionpy_controller.name_convention.NameClause'
                                      ''.format(self.nc_json_path, validator_name))

    def validate(self, name_to_validate):
        regex_res = self.regex.search(name_to_validate)
        if regex_res:
            return self._construct_parasble_result(name_to_validate, regex_res)
        else:
            return self._construct_failure_result(name_to_validate)

    def _construct_parasble_result(self, name_to_validate, regex_res):
        # If there is a regex result, then the name can be parsed
        result = {}
        for key in self._clause_validation:
            v = self._clause_validation[key]
            result[key] = v.validate(regex_res.group(key))
            valid = all(x.is_valid for x in result.values())

        class NamingResult(namedtuple(
                'NamingResult', self._clause_validation.keys())):
            __slots__ = ()

            @property
            def name_to_validate(self):
                return name_to_validate

            @property
            def is_parsable(self):
                return True

            @property
            def is_valid(self):
                return valid

            @property
            def get_message(self):
                if valid:
                    message = 'The name "{}" is parsable and valid:\n'.format(name_to_validate)
                else:
                    message = 'The name "{}" is parsable but not valid:\n'.format(name_to_validate)

                # map(lambda x: x.get_message, self._asdict().values())
                message = message + ('\n'.join(
                    ['\t{}'.format(x.get_message) for x in self._asdict().values()]
                ))
                return message

        return NamingResult(**result)

    def _construct_failure_result(self, name_to_validate):
        # Basic NamingResult for cases where the name cannot be parsed
        class NamingResult(namedtuple(
                'NamingResult', ('name_to_validate', 'is_parsable', 'is_valid', 'get_message'))):
            __slots__ = ()

        return NamingResult(
            name_to_validate,
            False, False, 'The name "{}" is not parsable'.format(name_to_validate)
        )


class NamingException(Exception):
    pass



import os.path
import six
from unittest import TestCase
from mapactionpy_controller.name_convention import NamingConvention, NamingException
from mapactionpy_controller.name_clause_validators import NamingClause, NamingLookupClause
from mapactionpy_controller.crash_move_folder import CrashMoveFolder

# works differently for python 2.7 and python 3.x
if six.PY2:
    import mock  # noqa: F401
else:
    from unittest import mock  # noqa: F401


class DummyClass(object):
    """
    Used within TestNamingConvention.test_load_dnc_definition() to test the case
    that a valid class is supplied, but one which doesn't inherit from
    mapactionpy_controller.name_convention.NameClause
    """

    def __init__(self, dummy_path, **kwargs):
        pass


class TestNamingConvention(TestCase):

    def setUp(self):
        parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

        self.dnc_json_path = os.path.join(
            parent_dir, 'example', 'data_naming_convention.json')

        self.test_files_dir = os.path.join(parent_dir, 'tests', 'testfiles', '.')

        cmf_descriptor_path = os.path.join(
            parent_dir, 'example', 'cmf_description.json')
        self.cmf = CrashMoveFolder(cmf_descriptor_path, verify_on_creation=False)
        self.cmf.dnc_lookup_dir = os.path.join(parent_dir, 'example')

    def test_load_csv_files_for_data_name_validator(self):
        # Test with a valid csv table
        dnlc = NamingLookupClause(self.dnc_json_path, filename='06_source.csv', name='test', lookup_field='Value')
        self.assertEqual(dnlc.lookup_field, 'Value')

        # Test with an valid csv file but a mismatch between primary key parameter and file contents
        self.assertRaises(NamingException, NamingLookupClause,
                          self.dnc_json_path, filename='06_source.csv', name='test',
                          lookup_field='non-existant-primary-key')

        # Test with an invalid csv table (duplicate primary key)
        self.assertRaises(NamingException, NamingLookupClause, self.test_files_dir,
                          filename='06_source_lookup_duplicate_prikey.csv',
                          name='test', lookup_field='Value')

    def test_load_dnc_definition(self):

        test_convention_files = (
            'fixture_name_convention_clause_def_and_regex_groupname_mismatch.json',
            'fixture_name_convention_clause_def_not_in_regex_groupname.json',
            'fixture_name_convention_missing_clause_def.json',
            'fixture_name_convention_incorrect_validation_class.json',
            'fixture_name_convention_nonexistant_validation_class.json'
        )

        for test_filename in test_convention_files:
            test_filepath = os.path.join(self.test_files_dir, test_filename)
            self.assertRaises(NamingException, NamingConvention, test_filepath)

    def test_abstract_validator(self):
        self.assertRaises(NotImplementedError, NamingClause, self.dnc_json_path)

        # Dummy implenmentation of calling the validate() method on NamingClause
        class DummyTestNameClause(NamingClause):
            def validate(self, clause_value, **kwargs):
                if six.PY2:
                    return super(DummyTestNameClause, self).validate(clause_value, **kwargs)
                else:
                    return super().validate(clause_value, **kwargs)

        tdnc = DummyTestNameClause(self.dnc_json_path)
        self.assertRaises(NotImplementedError, tdnc.validate, 'test')

    def test_get_other_dnc_attributes(self):
        # pylint: disable=no-member
        dnc = NamingConvention(self.dnc_json_path)
        # dnr.is_parsable is False if no match with regex
        dnr = dnc.validate(r'lka_admnad3_py_s0_wfpocha.pp.shp')
        self.assertFalse(dnr.is_parsable)

        # dnr.is_valid is False as not all clauses are found in lookups,
        # but details for the valid clauses are found.
        dnr = dnc.validate(r'aaa_admn_ad3_py_s0_wfp_pp')
        self.assertTrue(dnr.is_parsable)
        self.assertFalse(dnr.is_valid)
        self.assertFalse(dnr.geoext.is_valid)
        self.assertFalse(dnr.geoext.is_valid)
        self.assertTrue(dnr.datacat.is_valid)
        self.assertTrue(dnr.datatheme.is_valid)
        self.assertTrue(dnr.geom.is_valid)
        self.assertTrue(dnr.scale.is_valid)
        self.assertTrue(dnr.source.is_valid)
        self.assertTrue(dnr.perm.is_valid)
        self.assertTrue(dnr.freetext.is_valid)

        # clause details are found
        ref_datatheme = {'Value': 'ad3', 'Description': 'Administrative boundary (level 3)', 'Category': 'admn'}
        self.assertEqual(ref_datatheme, dnr.datatheme._asdict())
        # self.assertEqual(ref_datatheme, dnr.datatheme._asdict())

        ref_source = {'Value': 'wfp', 'Organisation': 'World Food Program', 'url': '',
                      'admn1Name': '', 'admn1PCode': '', 'admn2Name': '', 'admn2PCode': ''}
        self.assertEqual(ref_source, dnr.source._asdict())

        # with Free text clause present
        dnr = dnc.validate(r'aaa_admn_ad3_py_s0_wfp_pp_myfreetext')
        self.assertFalse(dnr.is_valid)
        # self.assertEqual(dnr.freetext.Value, 'myfreetext')

        # Fully valid name without Free text clause
        dnr = dnc.validate(r'lka_admn_ad3_py_s0_wfp_pp')
        self.assertTrue(dnr.is_valid)
        self.assertEqual(dnr.freetext.is_valid, True)
        self.assertIsNone(dnr.freetext.Value)

        # Fully valid name with Free text clause present
        dnr = dnc.validate(r'lka_admn_ad3_py_s0_wfp_pp_myfreetext')
        self.assertTrue(dnr.is_valid)
        self.assertEqual(dnr.freetext.Value, 'myfreetext')

    def test_name_validation(self):
        dnc = NamingConvention(self.dnc_json_path)
        # # pass valid names
        # dnr = dnc.validate(r'lka_admn_ad2_py_s5_unocha_pp_myfreetext')
        self.assertTrue(dnc.validate(r'lka_admn_ad2_py_s5_unocha_pp_freetext').is_parsable)
        self.assertTrue(dnc.validate(r'lka_admn_ad2_py_s5_unocha_pp_free_text').is_parsable)
        self.assertTrue(dnc.validate(r'lka_admn_ad3_py_s0_wfp_pp').is_parsable)
        # fail - fail regex
        self.assertFalse(dnc.validate(r'lka_admnad3_py_s0_wfpocha.pp.shp').is_parsable)
        # fail - clauses not listed in csv
        self.assertFalse(dnc.validate(r'aaa_admn_ad3_py_s0_wfp_pp').is_valid)
        self.assertFalse(dnc.validate(r'lka_bbbb_ad2_py_s5_ocha_pp_freetext').is_valid)
        self.assertFalse(dnc.validate(r'lka_admn_ccc_py_s5_ocha_pp_free_text').is_valid)
        # fail - clauses listed in wrong order
        dnr = dnc.validate(r'lka_admn_ad3_py_s0_wfpocha_pp')
        self.assertFalse(False if dnr is None else dnr.is_valid)
        dnr = dnc.validate(r'lka_ad3_admn_py_s0_wfpocha_pp')
        self.assertFalse(False if dnr is None else dnr.is_valid)
        dnr = dnc.validate(r'lka_admn_ad3_py_0s_pp_wfpocha')
        self.assertFalse(False if dnr is None else dnr.is_valid)

        # Clauses in upper case
        # Clauses with mixed case
        # Entire name in upper case
        self.assertTrue(dnc.validate(r'lKa_admn_ad3_py_s0_wfp_pp').is_valid)
        self.assertTrue(dnc.validate(r'LKA_admn_ad3_py_s0_wfp_pp').is_valid)
        self.assertTrue(dnc.validate(r'lka_aDmn_ad3_py_s0_wfp_pp').is_valid)
        self.assertTrue(dnc.validate(r'lka_ADMN_ad3_py_s0_wfp_pp').is_valid)
        self.assertTrue(dnc.validate(r'lka_admn_aD3_py_s0_wfp_pp').is_valid)
        self.assertTrue(dnc.validate(r'lka_admn_AD3_py_s0_wfp_pp').is_valid)
        self.assertTrue(dnc.validate(r'lka_admn_ad3_Py_s0_wfp_pp').is_valid)
        self.assertTrue(dnc.validate(r'lka_admn_ad3_PY_s0_wfp_pp').is_valid)
        self.assertTrue(dnc.validate(r'lka_admn_ad3_py_S0_wfp_pp').is_valid)
        self.assertTrue(dnc.validate(r'lka_admn_ad3_py_s0_wFp_pp').is_valid)
        self.assertTrue(dnc.validate(r'lka_admn_ad3_py_s0_WFP_pp').is_valid)
        self.assertTrue(dnc.validate(r'lka_admn_ad3_py_s0_wfp_pP').is_valid)
        self.assertTrue(dnc.validate(r'lka_admn_ad3_py_s0_wfp_PP').is_valid)
        self.assertTrue(dnc.validate(r'LKA_ADMN_AD3_PY_S0_WFP_PP').is_valid)

    def test_name_to_validate_property(self):
        dnc = NamingConvention(self.dnc_json_path)

        # test the full name is retrivable for a parsable name
        dnr = dnc.validate(r'aaa_admn_ad3_py_s0_wfp_pp')
        self.assertTrue(dnr.is_parsable)
        self.assertEqual(dnr.name_to_validate, r'aaa_admn_ad3_py_s0_wfp_pp')
        # test the full name is retrivable for a non-parsable name
        dnr = dnc.validate(r'lka_admnad3_py_s0_wfpocha.pp.shp')
        self.assertFalse(dnr.is_parsable)
        self.assertEqual(dnr.name_to_validate, r'lka_admnad3_py_s0_wfpocha.pp.shp')

    def test_get_message_property(self):
        dnc = NamingConvention(self.dnc_json_path)

        # message when dnr.is_parsable is False
        dnr = dnc.validate(r'lka_admnad3_py_s0_wfpocha.pp.shp')
        self.assertFalse(dnr.is_parsable)
        # print(dnr.get_message)
        self.assertRegexpMatches(dnr.get_message, r'not parsable')

        # message when dnr.is_parsable is True but dnr.is_valid is False
        dnr = dnc.validate(r'aaa_admn_ad3_py_s0_wfp_pp')
        self.assertTrue(dnr.is_parsable)
        self.assertFalse(dnr.is_valid)
        self.assertRegexpMatches(dnr.get_message, r'is parsable')
        self.assertRegexpMatches(dnr.get_message, r'not valid')
        # print(dnr.get_message)

        # message when dnr.is_valid is True
        dnr = dnc.validate(r'lka_admn_ad3_py_s0_wfp_pp')
        self.assertTrue(dnr.is_valid)
        self.assertRegexpMatches(dnr.get_message, r'is valid')
        # print(dnr.get_message)



def get_state_optional_fields(obj, optional_fields):
    # See https://docs.python.org/3/library/pickle.html#pickle-state
    # Copy the object's state from self.__dict__ which contains
    # all our instance attributes. Always use the dict.copy()
    # method to avoid modifying the original state.
    state = obj.__dict__.copy()
    # Remove any optional members which have a value of None.
    for option in optional_fields:
        if state[option] is None:
            del state[option]
    return state


def set_state_optional_fields(obj, state, optional_fields):
    # Restore optional elements to None.
    for option in optional_fields:
        if option not in state:
            state[option] = None

    obj.__dict__.update(state)



import fixtures
import os
import six
from unittest import TestCase
import json

from mapactionpy_controller.layer_properties import LayerProperties
from mapactionpy_controller.crash_move_folder import CrashMoveFolder
from mapactionpy_controller.map_cookbook import MapCookbook
from mapactionpy_controller.map_recipe import MapRecipe,  RecipeFrame
from mapactionpy_controller.recipe_layer import RecipeLayer


class TestMapCookBook(TestCase):

    def setUp(self):
        self.parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.path_to_valid_cmf_des = os.path.join(self.parent_dir, 'example', 'cmf_description_flat_test.json')
        self.path_to_invalid_cmf_des = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_cmf_description_one_file_and_one_dir_not_valid.json')

        self.cmf = CrashMoveFolder(self.path_to_valid_cmf_des, verify_on_creation=False)
        self.lyr_props = LayerProperties(self.cmf, '', verify_on_creation=False)

    def test_different_cmf_args(self):

        # 1) test with valid CMF and LayerProperties objects
        test_cmf = CrashMoveFolder(self.path_to_valid_cmf_des)
        test_lp = LayerProperties(test_cmf, "test", verify_on_creation=False)
        test_mcb = MapCookbook(test_cmf, test_lp, verify_on_creation=False)
        self.assertIsInstance(test_mcb, MapCookbook)

        # 2) test with invalid cmf object (eg and CrashMoveFolder object
        #    where verify_paths() returns False)
        test_cmf = CrashMoveFolder(self.path_to_invalid_cmf_des, verify_on_creation=False)
        self.assertRaises(ValueError, MapCookbook, test_cmf, test_lp)

    def test_layer_props_and_cmf_mismatch(self):

        # create and test_cmf and test_lp which refer different layer_prop.json files
        cmf1 = CrashMoveFolder(self.path_to_valid_cmf_des)
        test_lp = LayerProperties(cmf1, "test", verify_on_creation=False)
        # now point the test_cmf to a different layer_props.json
        cmf2 = CrashMoveFolder(self.path_to_valid_cmf_des)
        cmf2.layer_properties = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_layer_properties_four_layers.json'
        )

        # check that a ValueError is raised when `verify_on_creation=True`
        with self.assertRaises(ValueError) as ve:
            MapCookbook(cmf2, test_lp, verify_on_creation=True)

        if six.PY2:
            self.assertRegexpMatches(str(ve.exception), "strange results")
        else:
            self.assertRegex(str(ve.exception), "strange results")

        # check that a ValueError is not raised when `verify_on_creation=False`
        test_mcb = MapCookbook(cmf2, test_lp, verify_on_creation=False)
        self.assertIsInstance(test_mcb, MapCookbook)

    def test_layer_props_and_cookbook_mismatch(self):
        cmf = CrashMoveFolder(self.path_to_valid_cmf_des)
        cmf.layer_properties = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_layer_properties_four_layers.json'
        )
        cmf.map_definitions = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_cookbook_1map_4layers.json'
        )
        test_lp = LayerProperties(cmf, "test", verify_on_creation=False)

        # 1) map_cb and lp match
        test_mcb = MapCookbook(cmf, test_lp, verify_on_creation=True)
        self.assertIsInstance(test_mcb, MapCookbook)

        # 2) layer in map_cb but not in lp
        # 3) layer in lp but not in map_cb
        # 4) unmatched layers in both lp and map_cb
        testcases = (
            ('fixture_cookbook_1map_5layers.json', ("mainmap-tran-rds-ln-s0-allmaps")),
            ('fixture_cookbook_1map_3layers.json', ("locationmap_stle_stl_pt_s0_locationmaps")),
            ('fixture_cookbook_1map_mismatch_layers.json',
                ("mainmap-tran-rds-ln-s0-allmaps", "locationmap_stle_stl_pt_s0_locationmaps")
             ),
        )

        for cb_file, strings_in_ex_msg in testcases:
            cmf.map_definitions = os.path.join(self.parent_dir, 'tests', 'testfiles', cb_file)

            with self.assertRaises(ValueError) as ve:
                MapCookbook(cmf, test_lp, verify_on_creation=True)

            if six.PY2:
                self.assertRegexpMatches(str(ve.exception), "One or more layer names occur in only one of these files")
                for s in strings_in_ex_msg:
                    self.assertRegexpMatches(str(ve.exception), s)
            else:
                self.assertRegex(str(ve.exception), "One or more layer names occur in only one of these files")
                for s in strings_in_ex_msg:
                    self.assertRegex(str(ve.exception), s)

    def test_recipe_atlas_constructor(self):
        cmf = CrashMoveFolder(
            os.path.join(self.parent_dir, 'example', 'cmf_description_relative_paths_test.json'))
        cmf.layer_properties = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'cookbooks', 'fixture_layer_properties_for_atlas.json'
        )

        # 1) atlas matches other sections 'fixture_cookbook_good_with_atlas.json'
        cmf.map_definitions = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'cookbooks', 'fixture_cookbook_good_with_atlas.json'
        )
        test_lp = LayerProperties(cmf, ".lyr", verify_on_creation=False)
        MapCookbook(cmf, test_lp, verify_on_creation=True)
        self.assertTrue(True)

        # 2) mismatch map_frame name 'fixture_cookbook_atlas_mismatch_map_frame.json'
        # 3) mismatch layer name (but present in other map frame) 'fixture_cookbook_atlas_mismatch_layer1.json'
        # 4) mismatch layer name, not present elsewhere in the recipe. 'fixture_cookbook_atlas_mismatch_layer2.json'
        # 5) mismatch column_names
        test_cb_names = (
            'fixture_cookbook_atlas_mismatch_map_frame.json',
            'fixture_cookbook_atlas_mismatch_layer1.json',
            'fixture_cookbook_atlas_mismatch_layer2.json',
            'fixture_cookbook_atlas_mismatch_column_name.json'
        )

        for test_cb in test_cb_names:
            cmf.map_definitions = os.path.join(
                self.parent_dir, 'tests', 'testfiles', 'cookbooks', test_cb
            )
            test_lp = LayerProperties(cmf, ".lyr", verify_on_creation=False)
            with self.assertRaises(ValueError):
                MapCookbook(cmf, test_lp, verify_on_creation=True)

    def test_load_recipe_with_layer_props_inc(self):
        # Test that a MapRecipe read form json with only a layername, combined with the
        # relevant LayerProperties is equal to a MapRecipe with all of the layerdetails embeded.

        # layerpros with
        cmf = CrashMoveFolder(
            os.path.join(self.parent_dir, 'example', 'cmf_description_relative_paths_test.json'))
        cmf.layer_properties = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'cookbooks', 'fixture_layer_properties_for_atlas.json'
        )
        test_lp = LayerProperties(cmf, ".lyr", verify_on_creation=False)

        # recipe with layer name only
        recipe1 = MapRecipe(fixtures.recipe_with_layer_name_only, test_lp)
        # combined recipe with layer props
        recipe2 = MapRecipe(fixtures.recipe_with_layer_details_embedded, test_lp)

        self.assertEqual(recipe1, recipe2)

        # 2) with non-matching layer props schema
        recipe3 = MapRecipe(fixtures.recipe_with_positive_iso3_code, test_lp)
        self.assertNotEqual(recipe1, recipe3)

    def test_check_for_dup_text_elements(self):
        cmf = CrashMoveFolder(
            os.path.join(self.parent_dir, 'example', 'cmf_description_relative_paths_test.json'))
        cmf.layer_properties = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'cookbooks', 'fixture_layer_properties_for_atlas.json'
        )

        # 1) Pass with "good" text elements in just one map_frame
        cmf.map_definitions = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'cookbooks', 'fixture_cookbook_good_with_atlas.json'
        )
        test_lp = LayerProperties(cmf, ".lyr", verify_on_creation=False)
        MapCookbook(cmf, test_lp, verify_on_creation=True)
        self.assertTrue(True)

        # 2) Fail with duplicate text elements in multiple map_frames
        cmf.map_definitions = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'cookbooks', 'fixture_cookbook_with_dup_text_elements.json'
        )
        test_lp = LayerProperties(cmf, ".lyr", verify_on_creation=False)
        with self.assertRaises(ValueError) as ve:
            MapCookbook(cmf, test_lp, verify_on_creation=True)

        fail_msg = 'More than one "map_frame" is linked to the Scale text element "scale"'
        if six.PY2:
            self.assertRegexpMatches(str(ve.exception), fail_msg)
        else:
            self.assertRegex(str(ve.exception), fail_msg)

    def test_check_for_dup_layers_and_mapframs(self):
        cmf = CrashMoveFolder(
            os.path.join(self.parent_dir, 'example', 'cmf_description_relative_paths_test.json'))
        cmf.layer_properties = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'cookbooks', 'fixture_layer_properties_for_atlas.json'
        )

        # Fail with multiple layer with the same name in the same mapframe.
        test_cookbooks = (
            ('fixture_cookbook_with_dup_layers.json', 'mainmap_tran_por_pt_s0_allmaps'),
            ('fixture_cookbook_with_dup_mapframes.json', 'Main map')
        )

        for cb_filename, fail_msg in test_cookbooks:
            cmf.map_definitions = os.path.join(
                self.parent_dir, 'tests', 'testfiles', 'cookbooks', cb_filename
            )
            test_lp = LayerProperties(cmf, ".lyr", verify_on_creation=False)
            with self.assertRaises(ValueError) as ve:
                MapCookbook(cmf, test_lp, verify_on_creation=True)

            if six.PY2:
                self.assertRegexpMatches(str(ve.exception), fail_msg)
            else:
                self.assertRegex(str(ve.exception), fail_msg)

    def test_atlas_get_layer(self):
        recipe_def = json.loads(fixtures.recipe_with_positive_iso3_code)
        test_recipe = MapRecipe(recipe_def, self.lyr_props)
        atlas = test_recipe.get_frame('Main map')

        self.assertTrue(atlas.contains_layer('mainmap_stle_stl_pt_s0_allmaps'))
        self.assertFalse(atlas.contains_layer('not-existant'))

        lyr = atlas.get_layer('mainmap_stle_stl_pt_s0_allmaps')
        self.assertIsInstance(lyr, RecipeLayer)

        self.assertRaises(ValueError, atlas.get_layer, 'not-existant')

    def test_get_atlas(self):
        recipe_def = json.loads(fixtures.recipe_with_positive_iso3_code)
        test_recipe = MapRecipe(recipe_def, self.lyr_props)

        self.assertTrue(test_recipe.contains_frame('Main map'))
        self.assertFalse(test_recipe.contains_frame('not-existant'))

        lyr = test_recipe.get_frame('Main map')
        self.assertIsInstance(lyr, RecipeFrame)

        self.assertRaises(ValueError, test_recipe.get_frame, 'not-existant')

    def test_map_recipe_backward_compat(self):
        # This is required in order to get hold of a recipe object. This is because the method being
        # tested is a member of the class.
        recipe_obj = MapRecipe(fixtures.recipe_with_layer_name_only, self.lyr_props)

        test_cases = [
            (fixtures.recipe_schema_v2_0_with_layer_name_only, 0.2),
            (fixtures.recipe_with_layer_name_only, 0.3)
        ]

        for recipe_str, expected_result in test_cases:
            recipe_def = json.loads(recipe_str)
            self.assertEqual(recipe_obj._check_schemas_with_backward_compat(recipe_def), expected_result)

    def test_map_recipe_with_artbitary_principal_frame_name(self):
        # Test cases where the principal map frame is not called "Main map"
        # This should load without error
        MapRecipe(fixtures.recipe_with_non_standard_principal_map_frame_name, self.lyr_props)

        # This should raise a ValueError
        with self.assertRaises(ValueError):
            MapRecipe(fixtures.recipe_with_invalid_principal_map_frame_name, self.lyr_props)

    def test_filter_lyr_for_use_in_frame_extent(self):
        # Have included the digit at the start of the string, so that can be sorted easily.
        cmf = CrashMoveFolder(
            os.path.join(self.parent_dir, 'example', 'cmf_description_relative_paths_test.json'))
        cmf.layer_properties = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'cookbooks', 'fixture_layer_properties_for_atlas.json'
        )
        test_lp = LayerProperties(cmf, ".lyr", verify_on_creation=False)

        # recipe with layer name only
        with open(os.path.join(self.parent_dir, 'tests', 'testfiles',
                               'fixture_cookbook_1map_5layers_1frame.json')) as rf:
            cookbook_def = json.load(rf)
        # get the first (only) recipe in the cookbook
        recipe_def = cookbook_def['recipes'].pop()

        generic_lyr_def = json.loads('''{
            "name": "the_name",
            "reg_exp": "^wrl_admn_ad0_py_(.*?)_(.*?)_([phm][phm])(.+)shp$",
            "schema_definition": "admin1_reference.yml",
            "definition_query": "",
            "display": true,
            "add_to_legend": true,
            "label_classes": []
        }''')

        # Case 1
        # test white list
        test_white_list1 = [
            ('1a', True),
            ('1b', False),
            ('1c', None),
            ('1d', True),
            ('1e', False)
        ]
        expected_white_result1 = ['1a', '1d']

        test_white_list2 = [
            ('2a', True),
            ('2b', None),
            ('2c', None),
            ('2d', True),
            ('2e', None)
        ]
        expected_white_result2 = ['2a', '2d']

        # Case 2
        # Black List
        test_black_list = [
            ('3a', None),
            ('3b', False),
            ('3c', None),
            ('3d', None),
            ('3e', False)
        ]
        expected_black_result = ['3a', '3c', '3d']

        # Case 3
        # Default
        test_default_list = [
            ('4a', None),
            ('4b', None),
            ('4c', None),
            ('4d', None),
            ('4e', None)
        ]
        expected_default_result = ['4a', '4b', '4c', '4d', '4e']

        all_test_params = [
            (test_white_list1, expected_white_result1),
            (test_white_list2, expected_white_result2),
            (test_black_list, expected_black_result),
            (test_default_list, expected_default_result)
        ]

        for test_list, expected_result in all_test_params:
            test_recipe = MapRecipe(recipe_def, test_lp)

            # Build up a mock list of layer to test
            replacement_lyrs = []
            for test_lyr_details in test_list:
                new_lyr = RecipeLayer(generic_lyr_def, test_lp, verify_on_creation=False)
                new_lyr.name = test_lyr_details[0]
                new_lyr.use_for_frame_extent = test_lyr_details[1]
                # vaugely near Lebanon
                new_lyr.extent = (35, 33, 36, 34)
                new_lyr.crs = 'epsg:4326'

                replacement_lyrs.append(new_lyr)

            test_frame = test_recipe.map_frames.pop()
            test_frame.layers = replacement_lyrs
            result_lyrs = test_frame._filter_lyr_for_use_in_frame_extent()
            actual_result = [lyr.name for lyr in result_lyrs]
            self.assertEqual(actual_result, expected_result)

    def test_get_map_frame_extents(self):
        cmf = CrashMoveFolder(
            os.path.join(self.parent_dir, 'example', 'cmf_description_relative_paths_test.json'))
        cmf.layer_properties = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'cookbooks', 'fixture_layer_properties_for_atlas.json'
        )
        test_lp = LayerProperties(cmf, ".lyr", verify_on_creation=False)

        # recipe with layer name only
        with open(os.path.join(self.parent_dir, 'tests', 'testfiles',
                               'fixture_cookbook_1map_5layers_1frame.json')) as rf:
            cookbook_def = json.load(rf)
        # get the first (only) recipe in the cookbook
        recipe_def = cookbook_def['recipes'].pop()

        generic_lyr_def = json.loads('''{
            "name": "the_name",
            "reg_exp": "^wrl_admn_ad0_py_(.*?)_(.*?)_([phm][phm])(.+)shp$",
            "schema_definition": "admin1_reference.yml",
            "definition_query": "",
            "display": true,
            "add_to_legend": true,
            "label_classes": []
        }''')

        # Case 1
        # One or more layers does not have it's extent defined
        case1_list = [
            ('case1_lyrA', (35, 33, 36, 34), 'epsg:4326'),
            ('case1_lyrB', None, None)
        ]
        case1_result = (35, 33, 36, 34)

        # Case 2
        # Simple union with two lyrs of same crs
        case2_list = [
            ('case2_lyrA', (33, 51, 36, 58), 'epsg:4326'),
            ('case2_lyrB', (15, 52, 35, 55), 'epsg:4326')
        ]
        case2_result = (15, 51, 36, 58)

        # Case 3
        # Union with two lyrs of with different crs
        # 'epsg:4326'== WGS1984, 'epsg:3785' == Web Mercator
        case3_list = [
            ('case3_lyrA', (33, 51, 36, 58), 'epsg:4326'),
            ('case3_lyrB', (1669792.36, 6800125.45, 3896182.18, 7361866.11), 'epsg:3785')
        ]
        case3_result = (15, 51, 36, 58)

        # Case 4
        # One layer which stradles 180 degree meridian

        all_test_params = [
            (case1_list, case1_result),
            (case2_list, case2_result),
            (case3_list, case3_result)
        ]

        for test_list, expected_result in all_test_params:
            test_recipe = MapRecipe(recipe_def, test_lp)

            # Build up a mock list of layer to test
            replacement_lyrs = []
            for name, extent, crs in test_list:
                new_lyr = RecipeLayer(generic_lyr_def, test_lp, verify_on_creation=False)
                new_lyr.name = name
                new_lyr.use_for_frame_extent = bool(extent)
                # vaugely near Lebanon
                new_lyr.extent = extent
                new_lyr.crs = crs

                replacement_lyrs.append(new_lyr)

            # print('test_get_map_frame_extents')
            # print('expected_result = {}'.format(expected_result))
            test_frame = test_recipe.map_frames.pop()
            test_frame.crs = 'epsg:4326'
            test_frame.layers = replacement_lyrs
            test_frame.calc_extent(state=test_recipe)
            actual_result = test_frame.extent
            # print('actual_result = {}'.format(actual_result))
            for actual, expected in zip(actual_result, expected_result):
                self.assertAlmostEqual(actual, expected)
